#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from sqlalchemy import desc

from ..db import func_db, get_db


class DBStats(func_db.Model):
    __tablename__ = 'db_stats'
    __bind_key__ = 'logdb'

    id = func_db.Column(func_db.Integer, primary_key=True)
    type = func_db.Column(func_db.String(10))
    table = func_db.Column(func_db.String(50))
    period = func_db.Column(func_db.String(16))
    counter = func_db.Column(func_db.Float)

    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.current_timestamp())

    def __repr__(self):
        return '<dbstats %s>' % self.path

    @staticmethod
    def get_all():
        my_ret_val = get_db().session.query(DBStats).order_by(desc(DBStats.created)).all()

        return my_ret_val

    @staticmethod
    def get_since(a_date):
        my_ret_val = get_db().session.query(DBStats).filter(DBStats.created > a_date).order_by(desc(DBStats.created)).all()

        return my_ret_val




#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

from .cmd.root_cmd import RootCmd
from ..globals.job_mgr import JobMgr


class BusLogic:
    _job_mgr = None

    @staticmethod
    def get_instance():
        return BusLogic()

    def get_id(self):
        # this is default logic
        return 0

    def __init__(self):
        self._job_mgr = JobMgr()

    def get_job_mgr(self):
        return self._job_mgr

    def get_root_cmd(self):
        return RootCmd()




#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app
from flask_login import current_user
import uuid
import json
from ssk.lg import get_logic


class CmdHandler(object):
    PAGE_TERMINAL = 'ssk/admin/terminal.html'

    @staticmethod
    def get_terminal():
        return CmdHandler.PAGE_TERMINAL

    @staticmethod
    def cmd_post(a_request):
        my_request_json = a_request.json
        my_root_cmd = get_logic().get_root_cmd()
        try:
            my_params = [my_request_json["method"]]
            my_params.extend(my_request_json["params"])

            if current_user.is_admin():
                _, my_mesg = my_root_cmd.exec(my_params)
            else:
                my_mesg = "Access Denied"
        except Exception as an_e:
            my_error_id = str(uuid.uuid4())
            current_app.logger.error("{} {}".format(my_error_id, an_e))
            my_mesg = '[[ print "sorry, there was a problem, we will look into it asap.\nIt has been logged under: {}\n"]]'.format(
                my_error_id)

        my_response = {"jsonrpc": "2.0", "result": my_mesg, "id": 3}

        return json.dumps(my_response)



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app


from .abstract_cmd import AbstractCmd
from ...models.user import User


class ValidateUserCmd(AbstractCmd):
    def __init__(self):
        super().__init__("verify")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = ""

        if len(a_params) == 2:
            my_user = User.get_by_email(a_params[0])

            if my_user is not None:
                my_pass = a_params[1]
                user_manager = current_app.user_manager
                if user_manager.verify_password(my_pass, my_user.password):
                    my_ret_val = True
                    my_ret_mesg = '[[ print "OK: {} password match" ]]'.format(a_params[0])
                else:
                    my_ret_mesg = '[[ print "Error: {} password invalid" ]]'.format(a_params[0])
            else:
                my_ret_mesg = '[[ print "Error: user {} not found" ]]'.format(a_params[0])

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: verify <email> <password>\nverify user password" ]]'



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from ..db import func_db


class Status(func_db.Model):
    __tablename__ = 'status'
    __bind_key__ = 'logdb'

    id = func_db.Column(func_db.Integer, primary_key=True)
    api_status = func_db.Column(func_db.String(100))
    api_now_threshold = func_db.Column(func_db.String(100))
    api_active_now = func_db.Column(func_db.Integer)
    api_max_active = func_db.Column(func_db.Integer)
    api_total = func_db.Column(func_db.Integer)
    users = func_db.Column(func_db.Integer)
    users_active = func_db.Column(func_db.Integer)
    conf_keys = func_db.Column(func_db.Integer)
    audit_cnt = func_db.Column(func_db.Integer)
    response_time = func_db.Column(func_db.Integer)
    mem = func_db.Column(func_db.Float)

    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.current_timestamp())

    def __repr__(self):
        return '<Status %r>' % self.id



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

import time

from .base_job import BaseJob


class EmptyJob(BaseJob):
    def __init__(self, an_app, a_args):
        super(EmptyJob, self).__init__(an_app, a_args)

    def work(self):
        my_total = 60

        for my_tmp_idx in range(1, my_total):
            time.sleep(1)
            self.set_progress(a_val=round((my_tmp_idx / my_total) * 100))


#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

# from .base_job import BaseJob
# from ... import get_db
#
#
# class JupRenderJob(BaseJob):
#     def __init__(self, an_app, a_args):
#         super(JupRenderJob, self).__init__(an_app, a_args)
#
#     def work(self):
#         from ssk.models.db_stats import DBStats
#
#         with self._app.app_context():
#             try:
#                 my_all_tables = get_db().get_engine().table_names()
#                 my_log_tables = get_db().get_engine(bind_key="logdb").table_names()
#                 my_all_tables += my_log_tables
#
#                 for my_tmp_table in my_all_tables:
#                     my_count_sql = "select count(*) from {}".format(my_tmp_table)
#
#                     my_prefix = "func"
#                     if my_tmp_table not in my_log_tables:
#                         my_counter = get_db().engine.scalar(my_count_sql)
#                     else:
#                         my_prefix = "logdb"
#                         my_counter = get_db().get_engine(bind_key="logdb").scalar(my_count_sql)
#
#                     my_db_stat = DBStats()
#                     my_db_stat.type = "adhoc"
#                     my_db_stat.table = "{}.{}".format(my_prefix, my_tmp_table)
#                     my_db_stat.counter = my_counter
#
#                     my_db_session = get_db().session
#                     my_db_session.add(my_db_stat)
#                     my_db_session.commit()
#             except Exception as problem:
#                 self.write_to_log("Db Stat Failed {}".format(problem))



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from .. import db
from sqlalchemy.orm import relationship
from sqlalchemy import and_

from ..db import func_db


class Setting(func_db.Model):
    __tablename__ = 'setting'

    id = func_db.Column(func_db.Integer, primary_key=True)
    user_id = func_db.Column(func_db.Integer, func_db.ForeignKey('user.id'))
    key = func_db.Column(func_db.String(50), unique=False, nullable=False)
    type = func_db.Column(func_db.String(10), unique=False, nullable=False)
    value = func_db.Column(func_db.String(100), unique=False, nullable=False)
    system = func_db.Column('is_global', func_db.Boolean(), nullable=False, server_default='0')
    desc = func_db.Column(func_db.String(250), unique=False, nullable=True)
    updated = func_db.Column(func_db.DateTime(timezone=True), onupdate=func_db.func.current_timestamp())
    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.current_timestamp())

    owner = relationship("User", back_populates="user_settings")

    @staticmethod
    def get_by_owner_key(an_owners_id, a_key):
        my_retval = db.get_db().session.query(Setting).filter(and_(Setting.key == a_key,
                                                                   Setting.user_id == an_owners_id)).first()

        return my_retval

    @staticmethod
    def get_global_by_key(a_key):
        my_retval = db.get_db().session.query(Setting).filter(and_(Setting.key == a_key,
                                                                   Setting.system is True)).first()

        return my_retval

    @staticmethod
    def get_by_key(a_key):
        my_retval = db.get_db().session.query(Setting).filter(and_(Setting.key == a_key)).first()

        return my_retval

    def __repr__(self):
        return '<Setting %r>' % self.key



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


# Flask-WTF v0.13 renamed Flask to FlaskForm
try:
    from flask_wtf import FlaskForm             # Try Flask-WTF v0.13+
except ImportError:
    from flask_wtf import Form as FlaskForm     # Fallback to Flask-WTF v0.12 or older

from wtforms import BooleanField, HiddenField, PasswordField, SubmitField, StringField
from wtforms import validators

from .validators import unique_username_validator, unique_email_validator, password_validator


class EditUserProfileForm(FlaskForm):
    """Edit user profile form."""
    id = HiddenField('id')
    username = StringField('Username', validators=[validators.DataRequired(), unique_username_validator])
    first_name = StringField('First name', validators=[validators.DataRequired()])
    last_name = StringField('Last name', validators=[validators.DataRequired()])
    email = StringField('Email', validators=[validators.DataRequired(), validators.Email('Invalid Email'), unique_email_validator])
    active = BooleanField('Active')

    new_password = PasswordField('New Password', validators=[password_validator,
        ])
    retype_password = PasswordField('Retype New Password', validators=[
        validators.EqualTo('new_password', message='New Password and Retype Password did not match')
        ])

    submit = SubmitField('Save')



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

import os
from datetime import datetime

import psutil
import requests
from _datetime import timedelta
from flask import current_app

from .base_job import BaseJob
from ... import get_db
from ...globals.api_gate import ApiGate


class HealthCheckJob(BaseJob):
    def __init__(self, an_app, a_args):
        super(HealthCheckJob, self).__init__(an_app, a_args)

    def work(self):
        with self._app.app_context():
            try:
                from ...models.status import Status
                from ...models.user import User
                from ...models.audit import Audit
                from ...models.setting import Setting

                my_status = Status()

                my_process = psutil.Process(os.getpid())

                my_api_status = "open"
                if not ApiGate.is_open():
                    my_api_status = "closed"

                my_status.api_status = my_api_status
                my_status.api_now_threshold = current_app.config["API_ACTIVE_NOW"]
                my_status.api_active_now = ApiGate.num_active_keys()
                my_status.api_max_active = current_app.config["MAX_ACTIVE_KEYS"]
                my_status.api_total = ApiGate.num_total_keys()

                my_all_users = User.query.all()
                last_15min = datetime.now() - timedelta(minutes=15)
                my_loggedin_users = User.query.filter(User.last_access > last_15min).all()

                my_status.users = len(my_all_users)
                my_status.users_active = len(my_loggedin_users)

                my_all_settings = Setting.query.all()
                my_status.conf_keys = len(my_all_settings)

                my_all_audit = Audit.query.all()
                my_status.audit_cnt = len(my_all_audit)

                my_url = current_app.config["ROOT_URL"]

                my_resp_time = -1
                if my_url != "":
                    my_now = datetime.now()
                    my_response = requests.get(my_url)
                    if my_response.status_code == 200:
                        my_resp_time = (datetime.now() - my_now).total_seconds()
                else:
                    my_resp_time = 0

                my_status.response_time = my_resp_time
                my_status.mem = round(my_process.memory_info().rss / (1024 ** 2), 2)

                get_db().session.add(my_status)
                get_db().session.commit()
            except Exception as problem:
                self.write_to_log("Health Check Failed {}".format(problem))



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


import ssk.blueprints.home_handler
from ssk.forms.contact_form import ContactForm

from ssk.ssk_consts import SSK_ADMIN_GROUP
from ssk.globals.web_gate import WebGate

from flask import Blueprint, request
from flask import render_template

bp = Blueprint('local', __name__, url_prefix='/')


@bp.route('/', methods=['GET'])
def index():
    if WebGate.is_closed():
        return WebGate.render_closed()

    return render_template("local/start.html", admin_group_name=SSK_ADMIN_GROUP)


@bp.route('/about', methods=['GET'])
def about():
    if WebGate.is_closed():
        return WebGate.render_closed()

    my_form = ContactForm()
    my_form.source = "about"

    return render_template("local/about.html", form=my_form, message=None, admin_group_name=SSK_ADMIN_GROUP)


@bp.route('/post_contact', methods=['POST'])
def post_contact():
    if WebGate.is_closed():
        return WebGate.render_closed()

    _, my_form, my_message = ssk.blueprints.home_handler.HomeHandler.post_contact(request)

    return render_template("local/about.html", form=my_form, message=my_message, admin_group_name=SSK_ADMIN_GROUP)



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from ..db import func_db
from .. import db


class Contact(func_db.Model):
    __tablename__ = 'contact'

    id = func_db.Column(func_db.Integer, primary_key=True)
    req_id = func_db.Column(func_db.String(50), unique=False, nullable=False)
    source = func_db.Column(func_db.String(50), unique=False, nullable=False)
    name = func_db.Column(func_db.String(100), unique=False, nullable=False)
    email = func_db.Column(func_db.String(100), unique=False, nullable=False)
    details = func_db.Column(func_db.String(200), unique=False, nullable=False)
    sent = func_db.Column(func_db.Boolean, unique=False, nullable=True)

    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.current_timestamp())

    @staticmethod
    def get_by_key(a_key):
        my_ret_val = db.get_db().session.query(Contact).filter(Contact.req_id == a_key).first()

        return my_ret_val

    def __repr__(self):
        return '<Contact %r>' % self.post_id



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app
from wtforms import ValidationError


def password_validator(form, field):
    if field.data == "":
        return True

    current_app.user_manager.password_validator(form, field)


def unique_username_validator(form, field):
    my_valid_username = False
    my_new_name = field.data

    my_id = form.id.data

    from ..models.user import User
    my_curr_user = User.get_by_id(my_id)

    if my_curr_user.username == my_new_name:
        my_valid_username = True
    else:
        my_curr_user = User.get_by_name(my_new_name)

        if my_curr_user is None:
            my_valid_username = True

    if not my_valid_username:
        raise ValidationError('This Username is already in use. Please try another one.')


def unique_email_validator(form, field):
    my_valid_email = False
    my_new_email = field.data

    my_id = form.id.data

    from ..models.user import User
    my_curr_user = User.get_by_id(my_id)

    if my_curr_user.email == my_new_email:
        my_valid_email = True
    else:
        my_curr_user = User.get_by_email(my_new_email)

        if my_curr_user is None:
            my_valid_email = True

    if not my_valid_email:
        raise ValidationError('This Email is already in use. Please try another one.')


#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from sqlalchemy.orm import relationship

from .. import db
from ..db import func_db


class ApiKey(func_db.Model):
    __tablename__ = 'apikey'

    id = func_db.Column(func_db.Integer, primary_key=True)
    user_id = func_db.Column(func_db.Integer, func_db.ForeignKey('user.id'))
    active = func_db.Column('is_active', func_db.Boolean(), nullable=False, server_default='1')
    key = func_db.Column(func_db.String(50), unique=True, nullable=False)

    updated = func_db.Column(func_db.DateTime(timezone=True), onupdate=func_db.func.current_timestamp())
    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.current_timestamp())

    owner = relationship("User", back_populates="api_keys")

    @staticmethod
    def get_all_active():
        my_ret_val = db.get_db().session.query(ApiKey).filter(ApiKey.active == 1).all()

        return my_ret_val

    @staticmethod
    def get_by_key(a_key):
        my_ret_val = db.get_db().session.query(ApiKey).filter(ApiKey.key == a_key).first()

        return my_ret_val

    @staticmethod
    def get_by_user(a_user_id):
        my_ret_val = db.get_db().session.query(ApiKey).filter(ApiKey.user_id == a_user_id).all()

        return my_ret_val

    def __repr__(self):
        return '<ApiKey %r>' % self.id



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import g, current_app


def get_logic():
    if 'logic' not in g:
        g.logic = current_app.bus_logic.get_instance()

    return g.logic



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from datetime import datetime, timedelta

from statistics import mean

from ssk import AppSettings, get_db
from ssk.utils import now


class AdminHandler(object):
    PAGE_CLOSED = "ssk/closed.html"
    PAGE_SYSTEM_CHART = "ssk/system_chart.html"
    PAGE_SYSTEM_STATS = "ssk/system_stats.html"

    @staticmethod
    def system_chart():
        my_to_plot = {"labels": [], "datasets": [{}, {}, {}]}

        if AppSettings().get_setting("LOGIN_OFF"):
            return AdminHandler.PAGE_CLOSED, my_to_plot

        from ..models.status import Status
        my_now = datetime.now()
        my_threshold = my_now - timedelta(minutes=360)
        my_session = get_db().session
        my_records = my_session.query(Status).filter(Status.created > my_threshold).order_by(Status.id.asc()).all()

        my_times = [my_tmp_item.created for my_tmp_item in my_records]
        my_res_time = [my_tmp_item.response_time for my_tmp_item in my_records]
        my_mem = [my_tmp_item.mem for my_tmp_item in my_records]
        my_users_active = [my_tmp_item.users_active for my_tmp_item in my_records]

        my_res_time_ds = {"label": "Response Times",
                          "data": my_res_time,
                          "fill": False,
                          "borderColor": 'rgba(100, 153, 255 , 0.6)',
                          }

        my_mem_ds = {"label": "Memory",
                     "data": my_mem,
                     "fill": False,
                     "borderColor": 'rgba(204, 153, 255 , 0.6)',
                     }

        my_users_ds = {"label": "Users",
                       "data": my_users_active,
                       "fill": False,
                       "borderColor": 'rgb(201, 204, 63, 0.6)',
                       }

        my_to_plot = {"labels": my_times, "datasets": [my_res_time_ds, my_mem_ds, my_users_ds]}

        return AdminHandler.PAGE_SYSTEM_CHART, my_to_plot

    @staticmethod
    def system_stats():
        my_to_plot = {"labels": [], "datasets": [{}, {}, {}]}

        if AppSettings().get_setting("LOGIN_OFF"):
            return AdminHandler.PAGE_CLOSED, my_to_plot

        my_now = now()

        my_from = my_now - timedelta(days=21)
        from ssk.models.db_stats import DBStats
        my_recent_db_stats = DBStats.get_since(my_from)

        my_db_stats = {}
        my_all_db_tables = []
        for my_tmp_db_stat in my_recent_db_stats:
            my_day = my_tmp_db_stat.created.strftime("%Y-%m-%d")

            if my_day not in my_db_stats.keys():
                my_db_stats[my_day] = {}

            my_db_stats[my_day][my_tmp_db_stat.table] = my_tmp_db_stat.counter

            if my_tmp_db_stat.table not in my_all_db_tables:
                my_all_db_tables.append(my_tmp_db_stat.table)

        my_all_db_days = list(my_db_stats.keys())

        # page stats
        from ssk.models.stats import Stats
        my_page_stats = Stats.get_all()

        my_daily_stats = {}
        my_daily_perf = {}
        my_all_pages = []
        my_all_days = []

        my_per_day = {}

        for my_tmp_stat in my_page_stats:
            if my_tmp_stat.path not in my_all_pages:
                my_all_pages.append(my_tmp_stat.path)

            if my_tmp_stat.day not in my_per_day.keys():
                my_per_day[my_tmp_stat.day] = []

            my_per_day[my_tmp_stat.day].append(my_tmp_stat.mean_response_time)

            if my_tmp_stat.day not in my_all_days:
                my_all_days.append(my_tmp_stat.day)

            if my_tmp_stat.day not in my_daily_stats.keys():
                my_daily_stats[my_tmp_stat.day] = {}
                my_daily_perf[my_tmp_stat.day] = {}

            my_daily_stats[my_tmp_stat.day][my_tmp_stat.path] = my_tmp_stat.hits
            my_daily_perf[my_tmp_stat.day][my_tmp_stat.path] = round(my_tmp_stat.mean_response_time)

            if len(my_all_days) > 21:
                break

        my_days = list(my_per_day.keys())
        my_days.reverse()
        my_labels = my_days

        my_mean_rt = []

        for my_tmp_item in my_days:
            my_all_rts = my_per_day[my_tmp_item]
            if len(my_all_rts) > 0:
                my_mean_rt.append(mean(my_all_rts))
            else:
                my_mean_rt.append(0)

        my_res_time_ds = {"label": "Response Time",
                          "data": my_mean_rt,
                          "borderColor": 'rgba(100, 153, 255 , 0.6)',
                          }

        my_to_plot = {"labels": my_labels, "datasets": [my_res_time_ds]}

        my_all_days.reverse()

        my_ret_val = {"plot": my_to_plot,
                      "stats": my_daily_stats,
                      "perf": my_daily_perf,
                      "all_pages": my_all_pages,
                      "all_days": my_all_days,
                      "db_stats": my_db_stats,
                      "db_tables": my_all_db_tables,
                      "db_days": my_all_db_days}

        return AdminHandler.PAGE_SYSTEM_STATS, my_ret_val



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

import time
from logging.handlers import RotatingFileHandler

import yaml
from blinker import ANY
from flask.cli import with_appcontext
from flask import render_template, flash, request
from flask_login import current_user
from apscheduler.schedulers.background import BackgroundScheduler
from ssk.utils import set_const

import os
import click
from flask import current_app
import uuid

from .config import Config, get_log_level
from .globals.cmd_processor import CmdProcessor
from .globals.app_settings import AppSettings
from flask_login import logout_user
from flask_login.signals import user_logged_in, user_logged_out

from .lg import get_logic

from .db import get_db, truncate_password
from datetime import datetime

# Monkey-patch bcrypt.hashpw at the lowest level to handle 72-byte limit
# This must happen BEFORE passlib tries to use bcrypt
import bcrypt as _bcrypt_module
_original_bcrypt_hashpw = _bcrypt_module.hashpw
def _patched_bcrypt_hashpw(password, salt):
    """Patched bcrypt.hashpw that truncates passwords to 72 bytes."""
    if isinstance(password, str):
        password = password.encode('utf-8')
    if len(password) > 72:
        password = password[:72]
    return _original_bcrypt_hashpw(password, salt)
_bcrypt_module.hashpw = _patched_bcrypt_hashpw

# Also monkey-patch Flask-User's UserManager.hash_password
from flask_user import UserManager
_original_usermanager_hash_password = UserManager.hash_password
def _patched_hash_password(self, password):
    """Patched hash_password that truncates passwords to 72 bytes for bcrypt."""
    return _original_usermanager_hash_password(self, truncate_password(password))
UserManager.hash_password = _patched_hash_password

from .logic.jobs.db_cleanup_job import DbCleanupJob
from .logic.jobs.db_stat_job import DbStatJob
from .logic.jobs.health_check_job import HealthCheckJob
from .logic.jobs.page_stat_job import PageStatJob
from .models.audit import Audit
from .ssk_consts import SSK_ADMIN_GROUP

import logging
from flask import g


@user_logged_in.connect_via(ANY)
def user_logged_in(sender, user, **extra):
    if user.username != current_app.config["ADMIN_NAME"]:
        my_login_off = AppSettings().get_setting("LOGIN_OFF")
        if my_login_off is not None and my_login_off is True:
            logout_user()
            flash('Sorry, we are currently closed for the maintenance', 'error')
    else:
        user.last_login = datetime.now()

        if user.num_of_logins is not None:
            user.num_of_logins = user.num_of_logins + 1
        else:
            user.num_of_logins = 1

        user.loggedin = 1

        get_db().session.add(user)
        get_db().session.commit()


@user_logged_out.connect_via(ANY)
def user_logged_out(sender, user, **extra):
    user.loggedin = 0
    get_db().session.add(user)
    get_db().session.commit()


def request_error(e):
    my_error_id = str(uuid.uuid4())
    current_app.logger.error("{} {} {} {}".format(my_error_id, request.full_path, e.code, e.name, ))

    return render_template('error.html', error_id=my_error_id), 400


def internal_server_error(e):
    my_error_id = str(uuid.uuid4())
    current_app.logger.fatal("{} {}".format(my_error_id, e))

    return render_template('500.html', error_id=my_error_id), 500


def set_meta():
    from ssk.models.version import Version

    if "DB_CURR" not in current_app.config.keys() or "SSK_DB_CURR" not in current_app.config.keys():
        my_version_record = get_db().session.query(Version).order_by(Version.id.desc()).first()
        current_app.config["SSK_DB_CURR"] = my_version_record.ssk_version
        current_app.config["DB_CURR"] = my_version_record.db_version

    return {"meta": {"APP_NAME": current_app.config["USER_APP_NAME"],
                     "APP_PROFILE": current_app.config["PROFILE"],
                     "LG_VERSION": get_logic().get_id(),
                     "APP_VERSION": current_app.config["USER_APP_VERSION"],
                     "SSK_VERSION": current_app.config["SSK_VER"],
                     "DB_VERSION": current_app.config["DB_CURR"],
                     "SSK_DB_VERSION": current_app.config["SSK_DB_CURR"],
                     "LATEST_DB_VERSION": current_app.config["DB_MODEL_VERSION"],
                     "LATEST_SSK_DB_VERSION": current_app.config["SSK_DB_MODEL_VERSION"],
                     }}


def start_apigate():
    from .globals.api_gate import ApiGate
    ApiGate.init()
    ApiGate.load()


def start_settings():
    from .globals.app_settings import AppSettings
    from .globals.setting_parser import SettingParser

    my_admin_email = current_app.config["ADMIN_EMAIL"]
    my_value = current_app.config["WEBSITE_OPEN"]

    AppSettings().set_setting(my_admin_email, "WEBSITE_OPEN", SettingParser.BOOL_TYPE, my_value)


def after_request(response):
    if current_app.config["LOG_ALL_REQUESTS"]:
        my_response = int((time.time() - g.start) * 1000)
        if my_response > current_app.config["LOG_SLOWER_THAN"] or response.status_code >= 400:
            from ssk.models.access import Access
            my_access = Access()
            my_access.remote_addr = request.remote_addr
            my_access.method = request.method
            my_access.protocol = request.scheme
            my_access.path = request.full_path[:250]
            my_access.response = response.status

            my_access.response_time = my_response
            my_access.user = ""

            if not current_user.is_anonymous:
                my_access.user = current_user.username

            get_db().session.add(my_access)
            get_db().session.commit()

    return response


def every_request():
    if current_app.config["LOG_ALL_REQUESTS"]:
        g.start = time.time()

    my_user = current_user

    if not my_user.is_anonymous:
        if my_user.username != current_app.config["ADMIN_NAME"]:
            my_login_off = AppSettings().get_setting("WEBSITE_OPEN")
            if my_login_off is not None and my_login_off is False:
                logout_user()
                flash('Sorry, we are currently closed for the maintenance', 'error')
        else:
            from ssk.models.all_ssk_db import User
            my_user.last_access = datetime.now()

            get_db().session.add(my_user)
            get_db().session.commit()


def start_cmd_processor():
    my_max_collectors = current_app.config["MAX_COLLECTORS"]

    if my_max_collectors > 0:
        CmdProcessor.start(a_job_mgr=get_logic().get_job_mgr(),
                           a_logger=current_app.logger,
                           a_max_collectors=my_max_collectors)


def exec_cmd(a_cmd):
    with a_cmd.get_current_app().app_context():
        if AppSettings().get_setting("SCHED_ON"):
            a_cmd.write_to_audit("OK", "scheduler triggers")
            a_cmd.work()
        else:
            a_cmd.write_to_audit("NOK", "scheduler paused")


def start_scheduler():
    my_scheduler = BackgroundScheduler(daemon=True)
    my_minutes = 60
    current_app.logger.info("scheduling HealthCheckJob every {} minutes".format(my_minutes))
    my_health = HealthCheckJob(current_app, a_args=["health"])
    my_scheduler.add_job(exec_cmd, 'interval', minutes=my_minutes, args=[my_health])

    current_app.logger.info("scheduling PageStatJob every {} minutes".format(my_minutes))
    my_page_stat = PageStatJob(current_app, a_args=["stats"])
    my_scheduler.add_job(exec_cmd, 'interval', minutes=my_minutes, args=[my_page_stat])

    current_app.logger.info("scheduling DbStatJob every {} minutes".format(my_minutes))
    my_db_stat = DbStatJob(current_app, a_args=["dbstats"])
    my_scheduler.add_job(exec_cmd, 'interval', minutes=my_minutes, args=[my_db_stat])

    if "DB_CLEANUP" in current_app.config.keys():
        my_to_clean = current_app.config["DB_CLEANUP"]

        for my_tmp_table in my_to_clean.keys():
            current_app.logger.info("scheduling cleanup of {}".format(my_tmp_table))
            my_task = DbCleanupJob(current_app, a_args=["dbcleanup {}".format(my_tmp_table),
                                                        my_tmp_table,
                                                        my_to_clean[my_tmp_table]])

            my_scheduler.add_job(exec_cmd, 'interval', minutes=360, args=[my_task])

    my_scheduler.start()


def init_ssk(an_app, a_bus_logic, a_db_upgrader, a_testing):
    my_app = an_app
    my_app.db_upgrader = a_db_upgrader
    my_app.bus_logic = a_bus_logic

    flask_env = os.getenv("FLASK_ENV", None)
    my_profile = flask_env
    if a_testing:
        my_profile = "tst"

    my_app_config = os.getenv("SSK_CONFIG", None)
    if my_app_config is None:
        # Try multiple possible config locations
        possible_paths = [
            'app/cfg/{}.yaml'.format(my_profile),
            'cfg/{}.yaml'.format(my_profile),
        ]
        
        for path in possible_paths:
            if os.path.exists(path):
                my_app_config = path
                break
        
        if my_app_config is None:
            raise FileNotFoundError(
                f"Configuration file not found. Tried: {possible_paths}. "
                "Please set SSK_CONFIG environment variable to your config file path."
            )

    my_config_obj = Config()
    with open(my_app_config, 'r') as my_config:
        my_app_config = yaml.safe_load(stream=my_config)

        for my_tmp_key in my_app_config['ssk'].keys():
            my_type = list(my_app_config['ssk'][my_tmp_key].keys())[0]
            my_yaml_value = my_app_config['ssk'][my_tmp_key][my_type]
            my_value = set_const(my_tmp_key, my_yaml_value, my_type)
            my_config_obj.__setattr__(my_tmp_key, my_value)

    my_log_dir = my_config_obj.__getattribute__("LOG_DIR")
    if not os.path.exists(my_log_dir):
        os.mkdir(my_log_dir)

    my_config_obj.__setattr__("LOG_FILE", os.path.join(my_log_dir, my_config_obj.__getattribute__("LOG_FILE_NAME")))
    my_config_obj.__setattr__("SQLALCHEMY_BINDS", {'logdb': my_config_obj.__getattribute__("LOG_DB_CONNSTR")})

    my_app.config.from_object(my_config_obj)

    my_log_level = get_log_level(my_app)
    logging.basicConfig(handlers=[RotatingFileHandler(my_app.config["LOG_FILE"], maxBytes=5000000, backupCount=10)],
                        level=my_log_level,
                        format="[%(asctime)s] %(levelname)s in %(module)s: %(message)s")
    logging.getLogger('werkzeug').setLevel(my_log_level)

    my_app.logger.info("init_ssk {} start".format(ssk_consts.SSK_VER))

    if not os.path.exists(my_app.config["DATA_FOLDER"]):
        os.mkdir(my_app.config["DATA_FOLDER"])

    my_app.cli.add_command(clean_db_command)
    my_app.cli.add_command(create_db_command)
    my_app.cli.add_command(setup_db_command)
    my_app.cli.add_command(init_command)

    my_app.register_error_handler(500, internal_server_error)
    my_app.register_error_handler(400, request_error)
    my_app.register_error_handler(403, request_error)
    my_app.register_error_handler(404, request_error)
    my_app.register_error_handler(405, request_error)
    my_app.register_error_handler(410, request_error)

    from .db import close_db, close_logdb
    my_app.teardown_appcontext(close_db)
    my_app.teardown_appcontext(close_logdb)

    from .db import func_db
    func_db.init_app(my_app)

    my_app.before_request(every_request)
    my_app.after_request(after_request)

    my_app.context_processor(set_meta)

    from .models.user import User
    from .models.user import UserInvitation
    from flask_user import UserManager, EmailManager
    my_app.user_manager = UserManager(my_app, func_db, User, UserInvitationClass=UserInvitation)
    my_app.user_manager.email_manager = EmailManager(my_app)
    my_app.meta = {"PROFILE": os.getenv("FLASK_ENV", None)}

    from .blueprints import home
    my_app.register_blueprint(home.bp)

    from .blueprints import api
    my_app.register_blueprint(api.bp)

    from .blueprints import cmd
    my_app.register_blueprint(cmd.bp)

    from .blueprints import admin
    my_app.register_blueprint(admin.bp)

    my_app.logger.info("init_ssk end")

    return my_app


def start_ssk(an_app, a_testing):
    with an_app.app_context():
        an_app.logger.info("start_ssk start")

        with an_app.app_context():
            from .db import db_version_check
            db_version_check()

        if not a_testing:
            start_cmd_processor()
        start_apigate()
        # it has to be after db init
        start_scheduler()
        start_settings()

        an_app.logger.info("start_ssk end")


@click.command('clean-db')
@with_appcontext
def clean_db_command():
    """Clear the existing data and tables."""
    from . import db
    db.db_clean()
    click.echo('Database {} Removed'.format(current_app.config["DB_HOST"]))


@click.command('create-db')
@with_appcontext
def create_db_command():
    """Clear the existing data and create new tables."""
    from . import db
    db.db_create()
    click.echo('Database {} Created'.format(current_app.config["DB_HOST"]))


@click.command('setup-db')
@with_appcontext
def setup_db_command():
    """Setup initial config."""
    from . import db
    db.db_init()
    click.echo('Database {} Initialized'.format(current_app.config["DB_HOST"]))


@click.command('init')
@click.argument('app_name', required=False, default='app')
def init_command(app_name):
    """Initialize a new Soseki application with standard folder structure."""
    import os
    import shutil
    
    if os.path.exists(app_name):
        click.echo(f'Error: Directory "{app_name}" already exists')
        return
    
    # Get the path to the blanco_app template
    blanco_app_path = os.path.join(os.path.dirname(__file__), 'blanco_app')
    
    if not os.path.exists(blanco_app_path):
        click.echo(f'Error: Template directory not found at {blanco_app_path}')
        return
    
    # Copy the entire blanco_app directory
    try:
        shutil.copytree(blanco_app_path, app_name)
        click.echo(f'Created: {app_name}/ (from template)')
        
        # Get the parent directory where app_name is being created
        parent_dir = os.path.dirname(os.path.abspath(app_name))
        
        # Copy bin directory next to app folder
        blanco_bin_path = os.path.join(os.path.dirname(__file__), 'blanco_bin')
        bin_dest = os.path.join(parent_dir, 'bin')
        
        if os.path.exists(blanco_bin_path):
            if os.path.exists(bin_dest):
                shutil.rmtree(bin_dest)
            shutil.copytree(blanco_bin_path, bin_dest)
            click.echo(f'Created: bin/ (from template)')
        
        # Copy jup directory next to app folder
        blanco_jup_path = os.path.join(os.path.dirname(__file__), 'blanco_jup')
        jup_dest = os.path.join(parent_dir, 'jup')
        
        if os.path.exists(blanco_jup_path):
            if os.path.exists(jup_dest):
                shutil.rmtree(jup_dest)
            shutil.copytree(blanco_jup_path, jup_dest)
            click.echo(f'Created: jup/ (from template)')
        
        # List created structure
        for root, dirs, files in os.walk(app_name):
            level = root.replace(app_name, '').count(os.sep)
            indent = ' ' * 2 * level
            click.echo(f'{indent}{os.path.basename(root)}/')
            subindent = ' ' * 2 * (level + 1)
            for file in files:
                click.echo(f'{subindent}{file}')
        
        click.echo(f'\nSuccessfully initialized Soseki blanco application in "{app_name}/"')
        click.echo(f'\nNext steps:')
        click.echo(f'  # Edit cfg/lite.yaml with your settings')
        click.echo(f'  ./bin/run_app.sh')
        
    except Exception as e:
        click.echo(f'Error: Failed to create application: {e}')



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app
from sqlalchemy import desc
from ..utils import get_safe_string

from ..models.contact import Contact
import uuid


class ContactMgr:
    _db_session = None

    def __init__(self, a_db_session):
        self._db_session = a_db_session

    def commit(self):
        self._db_session.commit()

    def get_all_keys(self):
        my_db_list = list(self._db_session.query(Contact).order_by(desc(Contact.created)))

        my_ret_val = [my_tmp.req_id for my_tmp in my_db_list]

        return my_ret_val

    def add_post(self, a_name, an_email, a_message, a_source):
        my_ret_val = False
        my_ret_mesg = "Error: cannot add contact {}".format(a_name)

        my_req = self._db_session.query(Contact).filter(Contact.email == an_email).first()
        if my_req is None:
            try:
                my_request = Contact()
                my_request.req_id = str(uuid.uuid4())
                my_request.name = a_name
                my_request.email = an_email
                my_request.details = a_message
                my_request.source = a_source
                my_request.sent = False

                self._db_session.add(my_request)

                my_ret_val = True
                my_ret_mesg = '[[ print "OK: contact {} added" ]]'.format(my_request.req_id)
            except Exception:
                current_app.logger.error("Cannot add contact %s", get_safe_string(a_name))
        else:
            my_ret_mesg = '{} please be patient. We have your request already.'.format(my_req.email)

        return my_ret_val, my_ret_mesg



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

import tempfile
import os
import uuid
import json
from datetime import datetime
from abc import ABC, abstractmethod
from flask_login import current_user
from os.path import exists


class BaseJob(ABC):
    QUEUED = "QUEUED"
    IN_PROG_STATUS = "IN PROGRESS"
    STOPPED_STATUS = "STOPPED"
    DONE_STATUS = "DONE"

    _app = None
    _thread = None
    _task_id = None
    _args = None
    _progress = None
    _job_tracker = None
    _executor_id = None
    _logfile = None
    _status = None

    def __init__(self, an_app, an_args):
        self._app = an_app._get_current_object()

        if current_user._get_current_object() is not None and not current_user.is_anonymous:
            self._executor_id = current_user.id

        self._task_id = str(uuid.uuid4())
        self._args = an_args

    def set_executor_id(self, an_executor_id):
        self._executor_id = an_executor_id

    def get_current_app(self):
        return self._app

    def set_job_tracker(self, a_job_tracker):
        self._job_tracker = a_job_tracker

    def set_task_id(self, a_task_id):
        self._task_id = a_task_id

    def get_task_id(self):
        return self._task_id

    def get_args(self):
        return self._args

    def write_to_log(self, a_message):
        if self._logfile is not None and exists(self._logfile.name):
            my_now = datetime.now()
            my_message = "{} {}: {}\n".format(self._task_id,
                                              my_now.strftime("%Y-%m-%d %H:%M:%S"),
                                              a_message)
            self._logfile.write(my_message)
            self._logfile.flush()
        else:
            if self._logfile is None:
                self._app.logger.error("Task {} cannot log {}. Logfile name is not set".format(self._task_id,
                                                                                               a_message))
            else:
                if not exists(self._logfile.name):
                    self._app.logger.error("Task {} cannot log {}. Logfile does not exist".format(self._task_id,
                                                                                              a_message))

    def write_to_audit(self, a_status, a_message):
        with self._app.app_context():
            from ssk import Audit, get_db

            my_audit = Audit()
            my_audit.by_user = "SYSTEM"
            my_audit.category = "JOBS"
            my_audit.description = "{} job: {}".format(a_message, self.get_args())
            my_audit.status = a_status
            get_db().session.add(my_audit)
            get_db().session.commit()

    def get_args_str(self):
        try:
            my_ret_val = json.dumps(self._args)
        except Exception:
            my_ret_val = "Not Serializable args"

        return my_ret_val.replace("\"", "")

    def get_progress(self):
        return self._progress

    def get_logfile(self):
        if self._logfile is not None:
            return self._logfile.name

        return None

    def set_logfile(self, a_file):
        if a_file is not None and os.path.exists(a_file):
            self._logfile = open(a_file, 'a')

    def log_queue(self):
        self._status = BaseJob.QUEUED

        from ...models.job import Job

        with self._app.app_context():
            my_trigger = Job()
            my_trigger.task_id = self._task_id
            my_trigger.user_id = self._executor_id
            my_trigger.status = BaseJob.QUEUED
            my_trigger.progress = 0

            try:
                my_log_dir = self._app.config["LOG_DIR"]
                self._logfile = tempfile.NamedTemporaryFile(dir=my_log_dir, delete=False, mode="w")
                self.write_to_log("queued")

                my_trigger.logfile = self._logfile.name
            except FileNotFoundError:
                self._app.logger.error("Task {} cannot create logfile".format(self._task_id))

            if len(self._args) > 0:
                my_trigger.name = self._args[0]
            else:
                my_trigger.name = "noname"
            my_trigger.action = self.get_args_str()
            my_trigger.started = datetime.now()

            from ...db import get_db
            get_db().session.add(my_trigger)
            get_db().session.commit()

    def log_start(self):
        self.set_status(BaseJob.IN_PROG_STATUS)
        self.write_to_log("started")

    def log_done(self, an_app):
        with an_app.app_context():
            self.set_status(BaseJob.DONE_STATUS)

        self.write_to_log("done")

    def queue(self):
        self.log_queue()

    def start(self):
        self.log_start()
        self.execute(self._app)

    def stop(self):
        self.set_status(BaseJob.STOPPED_STATUS)

        if self._logfile is not None and exists(self._logfile.name):
            self.write_to_log("stopped")
            self._logfile.close()

    def done(self):
        self.set_status(BaseJob.DONE_STATUS)

        if self._logfile is not None and exists(self._logfile.name):
            self.write_to_log("done")
            self._logfile.close()

    def set_progress(self, a_val):
        self._progress = a_val

        with self._app.app_context():
            from ...models.job import Job
            from ...db import get_db
            my_trigger = get_db().session.query(Job).filter_by(task_id=self._task_id).first()

            if my_trigger is not None:
                my_trigger.progress = self._progress

                get_db().session.add(my_trigger)
                get_db().session.commit()

                if my_trigger.status == BaseJob.STOPPED_STATUS or my_trigger.status == BaseJob.DONE_STATUS:
                    self._app.logger.info("Task {} is {}".format(self._task_id, BaseJob.STOPPED_STATUS))
                    raise SystemExit()
            else:
                self._app.logger.info("Task {} not found".format(self._task_id))
                raise SystemExit()

    def execute(self, an_app):
        self.write_to_log("start execution")
        self.set_progress(0)
        try:
            self.work()
        except Exception as e:
            self.write_to_log("job execute error {}".format(e))
        finally:
            self.write_to_log("finally")

        self.log_done(an_app)
        self.set_progress(100)
        self._job_tracker.finish_job(self._task_id)

    def get_status(self):
        return self._status

    def set_status(self, a_status):
        self._status = a_status

        from ...models.job import Job
        from ...db import get_db

        with self._app.app_context():
            my_trigger = Job.get_by_key(self._task_id)

            if my_trigger is not None:
                my_trigger.status = a_status
                my_trigger.done = datetime.now()

                get_db().session.add(my_trigger)
                get_db().session.commit()
            else:
                self._app.logger.error("Task {} not found".format(self._task_id))
                raise SystemExit()

    @abstractmethod
    def work(self):
        ...



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import Blueprint, request

from ssk.blueprints.api_handler import ApiHandler

bp = Blueprint('api', __name__, url_prefix='/api')


@bp.route('/<string:a_version>/status/<string:a_key>', methods=['POST'])
def status(a_version, a_key):
    return ApiHandler.status(a_version, a_key)


@bp.route('/<string:a_version>/echo/<string:a_key>', methods=['POST'])
def echo(a_version, a_key):
    return ApiHandler.echo(request, a_version, a_key)




#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from ..db import func_db


class Audit(func_db.Model):
    __tablename__ = 'audit'

    id = func_db.Column(func_db.Integer, primary_key=True)
    by_user = func_db.Column(func_db.String(255), unique=False, nullable=True)
    category = func_db.Column(func_db.String(10), unique=False, nullable=True)
    description = func_db.Column(func_db.String(250), unique=False, nullable=False)
    status = func_db.Column(func_db.String(10), unique=False, nullable=True)
    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.current_timestamp())

    def __repr__(self):
        return '<Audit %r>' % self.description



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

import csv
import os

from flask import (Blueprint, request, Response, send_file, current_app)
from flask import render_template

from ssk.forms.contact_form import ContactForm
from ssk.ssk_consts import SSK_ADMIN_GROUP
from ssk.globals.web_gate import WebGate

from ssk.lg import get_logic

from ssk.blueprints.home_handler import HomeHandler


bp = Blueprint('home', __name__, url_prefix='/ssk')


@bp.route('/closed', methods=['GET'])
def closed():
    return WebGate.render_closed()


@bp.route('/', methods=['GET'])
def index():
    if WebGate.is_closed():
        return WebGate.render_closed()

    return render_template("local/start.html", admin_group_name=SSK_ADMIN_GROUP)


@bp.route('/tasks', methods=['GET'])
def tasks():
    if WebGate.is_closed():
        return WebGate.render_closed()

    my_file, my_tasks = HomeHandler.tasks()

    return render_template(my_file, tasks=my_tasks, admin_group_name=SSK_ADMIN_GROUP)


@bp.route('/blog/<string:note>', methods=['GET'])
def blog(note):
    if WebGate.is_closed():
        return WebGate.render_closed()

    my_note_path = "local/notes/{}.html".format(note)
    try:
        my_form = ContactForm()
        my_form.source = "note {}".format(note)

        return render_template("local/notebook.html",
                           toshow=my_note_path,
                           form=my_form,
                           message=None,
                           admin_group_name=SSK_ADMIN_GROUP)
    except Exception as err:
        current_app.logger.error(f"ERROR: note not found {my_note_path} {err}")

        return WebGate.render_notfound()


@bp.route('/blog_posts', methods=['GET'])
def blog_posts():
    if WebGate.is_closed():
        return WebGate.render_closed()

    my_template_path = current_app.jinja_loader.searchpath[0]
    my_meta_path = "{}/local/notes/meta.csv".format(my_template_path)

    my_posts = []
    with open(my_meta_path) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        line_count = 0
        for row in csv_reader:
            if line_count > 0:
                my_posts.append({"note": row[1], "title": row[2], "subtitle": row[3], "created": row[4], "labels": row[5]})
            line_count += 1

    return render_template("ssk/posts.html",
                           posts=my_posts,
                           admin_group_name=SSK_ADMIN_GROUP)


@bp.route('/tasks_action', methods=['POST'])
def tasks_action():
    _, my_tasks = HomeHandler.tasks_action(request)

    return render_template("ssk/tasks.html", tasks=my_tasks, admin_group_name=SSK_ADMIN_GROUP)


@bp.route('/task_log_download/<string:a_tid>')
def task_log_download(a_tid):
    my_logfile = get_logic().get_job_mgr().get_logfile(a_tid)

    if my_logfile is not None and os.path.exists(my_logfile):
        return send_file(my_logfile, as_attachment=True)


@bp.route('/progress/<string:a_thread_id>')
def progress(a_thread_id):
    def generate(a_job_tracker, a_tid):
        my_progress = a_job_tracker.get_progress(a_tid)
        yield "data:" + str(my_progress) + "\n\n"

    return Response(generate(get_logic().get_job_mgr(), a_thread_id), mimetype='text/event-stream')


@bp.route('/about', methods=['GET'])
def about():
    if WebGate.is_closed():
        return WebGate.render_closed()

    my_form = ContactForm()
    my_form.source = "about"

    return render_template("ssk/about.html", form=my_form, message=None, admin_group_name=SSK_ADMIN_GROUP)


@bp.route('/post_contact', methods=['POST'])
def post_contact():
    if WebGate.is_closed():
        return WebGate.render_closed()

    my_page, my_form, my_message = HomeHandler.post_contact(request)

    return render_template(my_page, form=my_form, message=my_message, admin_group_name=SSK_ADMIN_GROUP)



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app
from sqlalchemy import desc

from .abstract_cmd import AbstractCmd
from ...models.user import User
from ...utils import get_padding, get_ago
from ...db import get_db
from datetime import datetime


class UserDisableCmd(AbstractCmd):
    def __init__(self):
        super().__init__("disable")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = self.help()

        if len(a_params) == 1:
            if a_params[0] == "all":
                my_users = get_db().session.query(User).filter(User.username != current_app.config["ADMIN_NAME"]).all()

                for my_tmp_user in my_users:
                    my_tmp_user.active = False
                    get_db().session.add(my_tmp_user)

                get_db().session.commit()

                my_ret_val = True
                my_ret_mesg = '[[ print "OK: all non admin users disabled" ]]'
            else:
                my_user = User.get_by_email(a_params[0])

                if my_user is not None:
                    my_user.active = False

                    get_db().session.add(my_user)
                    get_db().session.commit()

                    my_ret_val = True
                    my_ret_mesg = '[[ print "OK: user {} disabled" ]]'.format(a_params[0])
                else:
                    my_ret_mesg = UserCmd.ERROR_USER_NOT_FOUND.format(a_params[0])

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: users off [<email>|all]\ndisable a user" ]]'


class UserRemoveCmd(AbstractCmd):
    def __init__(self):
        super().__init__("del")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = self.help()

        if len(a_params) == 1:
            my_user = User.get_by_email(a_params[0])

            if my_user is not None:
                if not my_user.active:
                    get_db().session.delete(my_user)
                    get_db().session.commit()

                    my_ret_val = True
                    my_ret_mesg = '[[ print "OK: user {} removed" ]]'.format(a_params[0])
                else:
                    my_ret_mesg = '[[ print "Error: Cannot remove active user {}" ]]'.format(a_params[0])
            else:
                my_ret_mesg = UserCmd.ERROR_USER_NOT_FOUND.format(a_params[0])

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: users del <email>\nremove a user" ]]'


class UserEnableCmd(AbstractCmd):
    def __init__(self):
        super().__init__("enable")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = self.help()

        if len(a_params) == 1:
            if a_params[0] == "all":
                my_users = get_db().session.query(User).filter(User.username != current_app.config["ADMIN_NAME"]).all()

                for my_tmp_user in my_users:
                    my_tmp_user.active = True
                    get_db().session.add(my_tmp_user)

                get_db().session.commit()

                my_ret_val = True
                my_ret_mesg = '[[ print "OK: all non admin users enabled" ]]'
            else:
                my_user = User.get_by_email(a_params[0])
                if my_user is not None:
                    my_user.active = True

                    get_db().session.add(my_user)
                    get_db().session.commit()

                    my_ret_val = True
                    my_ret_mesg = '[[ print "OK: user {} enabled" ]]'.format(a_params[0])
                else:
                    my_ret_mesg = UserCmd.ERROR_USER_NOT_FOUND.format(a_params[0])

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: users on [<email>|all]\nenable a user" ]]'


class UserPrintCmd(AbstractCmd):
    def __init__(self):
        super().__init__("list")

    def action(self, a_param: list):
        my_mesg = '[[ print "\n'

        my_template = "{} {} {} {} {} {} {} {} {} {} {}\n"
        my_users = User.query.order_by(desc(User.created)).all()
        my_mesg = my_mesg + my_template.format(get_padding("id", 6),
                                               get_padding("username", 10),
                                               get_padding("groups", 20),
                                               get_padding("subs", 10),
                                               get_padding("on", 4),
                                               get_padding("email", 30),
                                               get_padding("last login", 12),
                                               get_padding("active", 12),
                                               get_padding("#", 4),
                                               get_padding("registered", 20),
                                               get_padding("confirmed", 30))

        my_mesg = my_mesg + my_template.format("_" * 6,
                                               "_" * 10,
                                               "_" * 20,
                                               "_" * 10,
                                               "_" * 4,
                                               "_" * 30,
                                               "_" * 12,
                                               "_" * 12,
                                               "_" * 4,
                                               "_" * 20,
                                               "_" * 30)

        for my_tmp_user in my_users:
            my_roles = (myTmpRole.name for myTmpRole in my_tmp_user.roles)
            my_roles_str = ", ".join(my_roles)

            my_subs = (myTmpSub.name for myTmpSub in my_tmp_user.subs)
            my_subs_str = ", ".join(my_subs)

            my_on = "ON"
            if not my_tmp_user.active:
                my_on = "OFF"

            my_last_login_str = "-"
            if my_tmp_user.last_login is not None:
                my_last_login_str = get_ago((datetime.now() - my_tmp_user.last_login).seconds)

            my_last_access_str = "-"
            if my_tmp_user.last_access is not None:
                my_last_access_str = get_ago((datetime.now() - my_tmp_user.last_access).seconds)

            confirmed_str = "-"
            if my_tmp_user.email_confirmed_at is not None:
                confirmed_str = get_ago((my_tmp_user.email_confirmed_at - my_tmp_user.created).seconds,
                                        a_suffix="later")

            my_mesg = my_mesg + my_template.format(get_padding(my_tmp_user.id, 6),
                                                   get_padding(my_tmp_user.username, 10),
                                                   get_padding(my_roles_str, 20),
                                                   get_padding(my_subs_str, 10),
                                                   get_padding(my_on, 4),
                                                   get_padding(my_tmp_user.email, 30),
                                                   get_padding(my_last_login_str, 12),
                                                   get_padding(my_last_access_str, 12),
                                                   get_padding(my_tmp_user.num_of_logins, 4),
                                                   get_padding(my_tmp_user.created, 20),
                                                   get_padding(confirmed_str, 30))

        my_mesg = my_mesg + '" ]]'

        return True, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: user ls\nprints the list of all users" ]]'


class UserCmd(AbstractCmd):
    ERROR_USER_NOT_FOUND = '[[ print "Error: User {} not found" ]]'

    def __init__(self):
        super().__init__("user")
        self.reg_cmd(["l", "list"], UserPrintCmd())
        self.reg_cmd(["e", "enable"], UserEnableCmd())
        self.reg_cmd(["d", "disable"], UserDisableCmd())
        self.reg_cmd(["r", "remove"], UserRemoveCmd())

    def action(self, a_params: list):
        my_result = False
        my_mesg = self.help()

        my_cmd_name = self.get_params(a_params)
        my_cmd = self.get_cmd(my_cmd_name)
        if my_cmd is not None:
            my_result, my_mesg = my_cmd.exec(a_params)

        return my_result, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: user {list | enable | disable | remove}" ]]'



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

from ssk.models.audit import Audit
from ssk.models.user import User, Role, UserRoles, UserInvitation
from ssk.models.version import Version
from ssk.models.job import Job
from ssk.models.setting import Setting
from ssk.models.apikey import ApiKey



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from .. import db
from ..db import func_db
from sqlalchemy.orm import relationship, exc
from flask_user import UserMixin
from ..ssk_consts import SSK_ADMIN_GROUP


class User(func_db.Model, UserMixin):
    USER_ID = 'user.id'

    __tablename__ = 'user'

    id = func_db.Column(func_db.Integer, primary_key=True)
    username = func_db.Column(func_db.String(80), unique=True, nullable=True)

    active = func_db.Column('is_active', func_db.Boolean(), nullable=False, server_default='1')
    email = func_db.Column(func_db.String(255), nullable=True, unique=True)
    email_confirmed_at = func_db.Column('confirmed_at', func_db.DateTime())

    loggedin = func_db.Column('loggedin', func_db.Boolean(), nullable=False, server_default='0')

    password = func_db.Column(func_db.String(255), unique=False, nullable=False)
    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.current_timestamp())
    updated = func_db.Column(func_db.DateTime(timezone=True), onupdate=func_db.func.current_timestamp())
    last_login = func_db.Column(func_db.DateTime(timezone=True))
    last_access = func_db.Column(func_db.DateTime(timezone=True))
    num_of_logins = func_db.Column(func_db.Integer, nullable=False, server_default='0')

    first_name = func_db.Column(func_db.String(50), nullable=True)
    last_name = func_db.Column(func_db.String(50), nullable=True)

    roles = relationship('Role', secondary='user_roles', back_populates="users")
    #version 8
    subs = relationship('Subs', secondary='user_subs', back_populates="users")

    jobs = relationship("Job", back_populates="executor")
    user_settings = relationship("Setting", back_populates="owner")
    api_keys = relationship("ApiKey", back_populates="owner")

    def __repr__(self):
        return self.email

    def has_role(self, a_role):
        my_role_names = (myTmpRole.name for myTmpRole in self.roles)

        return a_role in my_role_names

    def is_admin(self):
        my_role_names = (myTmpRole.name for myTmpRole in self.roles)

        return SSK_ADMIN_GROUP in my_role_names

    @staticmethod
    def all_roles_str(an_inid):
        my_user = db.get_db().session.query(User).filter(User.inid == an_inid).first()
        my_ret_val = (myTmpRole.name for myTmpRole in my_user.roles)

        return ", ".join(my_ret_val)

    @staticmethod
    def get_by_inid(an_inid):
        my_ret_val = db.get_db().session.query(User).filter(User.inid == an_inid).first()

        return my_ret_val

    @staticmethod
    def get_by_id(an_id):
        my_ret_val = db.get_db().session.query(User).filter(User.id == an_id).first()

        return my_ret_val

    @staticmethod
    def get_by_email(an_email):
        my_ret_val = db.get_db().session.query(User).filter(User.email == an_email).first()

        return my_ret_val

    @staticmethod
    def get_by_name(a_name):
        my_ret_val = None
        try:
            my_ret_val = db.get_db().session.query(User).filter(User.username == a_name).one()
        except exc.NoResultFound:
            print("User {} not found".format(a_name))

        return my_ret_val


class Role(func_db.Model):
    ROLE_ID = "role.id"
    __tablename__ = 'role'

    id = func_db.Column(func_db.Integer(), primary_key=True)
    name = func_db.Column(func_db.String(50), unique=True)

    active = func_db.Column('is_active', func_db.Boolean(), nullable=True, server_default='1')

    users = relationship('User', secondary='user_roles', back_populates="roles")

    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.now())
    updated = func_db.Column(func_db.DateTime(timezone=True), onupdate=func_db.func.now())

    @staticmethod
    def get_by_name(a_name):
        return db.get_db().session.query(Role).filter(Role.name == a_name).first()

class UserRoles(func_db.Model):
    __tablename__ = 'user_roles'

    user_id = func_db.Column(func_db.Integer(), func_db.ForeignKey(User.USER_ID), primary_key=True)
    role_id = func_db.Column(func_db.Integer(), func_db.ForeignKey(Role.ROLE_ID), primary_key=True)

    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.now())
    updated = func_db.Column(func_db.DateTime(timezone=True), onupdate=func_db.func.now())


class UserInvitation(func_db.Model):
    __tablename__ = 'user_invitations'
    id = func_db.Column(func_db.Integer, primary_key=True)
    email = func_db.Column(func_db.String(255), nullable=False)
    invited_by_user_id = func_db.Column(func_db.Integer, func_db.ForeignKey(User.USER_ID))

#version 8
class Transaction(func_db.Model):
    __tablename__ = 'user_transaction'
    id = func_db.Column(func_db.Integer, primary_key=True)
    description = func_db.Column(func_db.String(255), nullable=False)
    invited_by_user_id = func_db.Column(func_db.Integer, func_db.ForeignKey(User.USER_ID))


class Subs(func_db.Model):
    SUBS_ID = "subs.id"
    __tablename__ = 'subs'

    id = func_db.Column(func_db.Integer(), primary_key=True)
    name = func_db.Column(func_db.String(50), unique=True)

    active = func_db.Column('is_active', func_db.Boolean(), nullable=True, server_default='1')
    price = func_db.Column(func_db.Float())

    users = relationship('User', secondary='user_subs', back_populates="subs")

    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.now())
    updated = func_db.Column(func_db.DateTime(timezone=True), onupdate=func_db.func.now())

    @staticmethod
    def get_by_name(a_name):
        return db.get_db().session.query(Subs).filter(Subs.name == a_name).first()


class UserSubs(func_db.Model):
    __tablename__ = 'user_subs'

    user_id = func_db.Column(func_db.Integer(), func_db.ForeignKey(User.USER_ID), primary_key=True)
    subs_id = func_db.Column(func_db.Integer(), func_db.ForeignKey(Subs.SUBS_ID), primary_key=True)

    expiration = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.now())
    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.now())
    updated = func_db.Column(func_db.DateTime(timezone=True), onupdate=func_db.func.now())



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


import os

from flask import current_app
from sqlalchemy.orm import exc
from sqlalchemy import and_, desc

from ..logic.jobs.base_job import BaseJob
from ..logic.jobs.empty_job import EmptyJob
from ..models.job import Job
from ..db import get_db


class JobMgr:
    _queued_jobs = None
    _active_jobs = None

    def __init__(self):
        # every thread has new g object, meaning also new JobMgr
        # in case of a progress request from the gui, new thread needs to load jobs from the db
        # to be able to monitor status
        self._active_jobs = {}

        my_all_jobs = get_db().session.query(Job).filter(Job.status == BaseJob.IN_PROG_STATUS).order_by(desc(Job.created)).all()

        for my_tmp_job in my_all_jobs:
            my_tmp_task = EmptyJob(current_app, a_args=my_tmp_job.action)
            my_tmp_task.set_executor_id(my_tmp_job.user_id)
            my_tmp_task.set_task_id(my_tmp_job.task_id)
            my_tmp_task.set_progress(my_tmp_job.progress)
            my_tmp_task.set_logfile(my_tmp_job.logfile)

            my_tmp_task.set_job_tracker(self)
            self._active_jobs[my_tmp_task.get_task_id()] = my_tmp_task

        self._queued_jobs = {}

        my_all_jobs = get_db().session.query(Job).filter(Job.status == BaseJob.QUEUED).order_by(desc(Job.created)).all()

        for my_tmp_job in my_all_jobs:
            my_tmp_task = EmptyJob(current_app, a_args=my_tmp_job.action)
            my_tmp_task.set_executor_id(my_tmp_job.user_id)
            my_tmp_task.set_task_id(my_tmp_job.task_id)
            my_tmp_task.set_progress(my_tmp_job.progress)
            my_tmp_task.set_logfile(my_tmp_job.logfile)

            my_tmp_task.set_job_tracker(self)
            self._queued_jobs[my_tmp_task.get_task_id()] = my_tmp_task

    def total_active(self):
        return len(self._active_jobs)

    def total_queued(self):
        return len(self._queued_jobs)

    def is_active(self, a_job_id):
        return a_job_id in self._active_jobs.keys()

    def is_queued(self, a_job_id):
        return a_job_id in self._queued_jobs.keys()

    def queue_job(self, a_job):
        a_job.set_job_tracker(self)
        self._queued_jobs[a_job.get_task_id()] = a_job
        a_job.queue()

        return a_job

    def start_job(self, a_job):
        if a_job.get_task_id() in self._queued_jobs.keys():
            self._queued_jobs.pop(a_job.get_task_id())

        self._active_jobs[a_job.get_task_id()] = a_job
        a_job.start()

        return a_job

    def get_last_active_job_by_user(self, a_user_id, a_name=None):
        if a_name is not None:
            return get_db().session.query(Job).filter(and_(Job.user_id == a_user_id,
                                                           Job.status == BaseJob.IN_PROG_STATUS,
                                                           Job.name == a_name)).order_by(desc(Job.created)).first()
        else:
            return get_db().session.query(Job).filter(and_(Job.user_id == a_user_id,
                                                           Job.status == BaseJob.IN_PROG_STATUS)).order_by(desc(Job.created)).first()

    def get_all_active_jobs_by_user(self, a_user_id):
        my_all_jobs = get_db().session.query(Job).filter(and_(Job.user_id == a_user_id, Job.status == BaseJob.IN_PROG_STATUS)).order_by(desc(Job.created)).all()

        return my_all_jobs

    def get_all_jobs_by_user(self, a_user_id):
        my_all_jobs = get_db().session.query(Job).filter(and_(Job.user_id == a_user_id)).order_by(desc(Job.created)).all()

        return my_all_jobs

    def get_progress(self, a_job_id):
        my_ret_val = 0

        if a_job_id != "-":
            my_ret_val = 100

            if self.is_active(a_job_id):
                my_job = self._active_jobs[a_job_id]
                my_ret_val = my_job.get_progress()

        return my_ret_val

    def get_logfile(self, a_job_id):
        my_retval = None
        my_job = Job.get_by_key(a_job_id)

        if my_job is not None and os.path.exists(my_job.logfile):
            my_retval = my_job.logfile

        return my_retval

    def stop_job(self, a_job_id):
        my_ret_val = False
        my_mesg = ""

        if self.is_active(a_job_id):
            my_job = self._active_jobs[a_job_id]
            my_job.stop()

            self._active_jobs.pop(a_job_id)

            my_ret_val = True
        elif self.is_queued(a_job_id):
            my_job = self._queued_jobs[a_job_id]
            my_job.stop()

            self._queued_jobs.pop(a_job_id)

            my_ret_val = True
        else:
            my_mesg = "Error: Job {} Not Found".format(a_job_id)

        return my_ret_val, my_mesg

    def finish_job(self, a_job_id):
        my_ret_val = False
        if self.is_active(a_job_id):
            self._active_jobs.pop(a_job_id)

            my_ret_val = True

        return my_ret_val

    def stop_all(self, a_status):
        my_to_stop = {}
        if a_status == BaseJob.IN_PROG_STATUS:
            my_to_stop = dict(self._active_jobs)
        if a_status == BaseJob.QUEUED:
            my_to_stop = dict(self._queued_jobs)

        my_cnt = 0
        for my_job_id in my_to_stop:
            my_res, _ = self.stop_job(my_job_id)
            if my_res:
                my_cnt += 1

        return True, "Stopped {} jobs".format(my_cnt)

    def del_all(self, a_status):
        my_cnt = 0
        my_all = Job.query.filter_by(status=a_status).all()
        for my_tmp_job in my_all:
            my_res, _ = self.delete_job(my_tmp_job.task_id)
            if my_res:
                my_cnt += 1

        return True, "Deleted {} jobs".format(my_cnt)

    def delete_job(self, a_job_id):
        my_retval = False

        from ..models.job import Job
        from ..db import get_db

        try:
            if not self.is_active(a_job_id) and not self.is_queued(a_job_id):
                my_job = Job.get_by_key(a_job_id)
                if my_job is not None:
                    if my_job.logfile is not None and os.path.exists(my_job.logfile):
                        try:
                            os.remove(my_job.logfile)
                        except Exception as e:
                            current_app.logger.error("Task {} cannot remove logfile {} {}".format(a_job_id,
                                                                                                  my_job.logfile,
                                                                                                  e))
                    else:
                        current_app.logger.error("Task {} logfile {} does not exist".format(a_job_id, my_job.logfile))

                    Job.query.filter_by(task_id=a_job_id).delete()
                    get_db().session.commit()

                my_retval = True
                my_mesg = "OK: Job {} deleted".format(a_job_id)
            else:
                my_mesg = "Error: Cannot delete active or queued job {} ".format(a_job_id)

        except exc.NoResultFound:
            my_mesg = "Error: Job {} not found".format(a_job_id)

        return my_retval, my_mesg



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

from datetime import datetime

from flask import current_app
from flask_user import user_manager
from sqlalchemy import create_engine

from ssk.db import get_db, get_version, set_ssk_version, func_db
from ssk.globals.setting_parser import SettingParser
from ssk.models.setting import Setting
from ssk.models.user import User


class SSKUpgrader:
    _to_skip = []
    UPGRADING_MESG = "upgrading to {}"

    @staticmethod
    def ver1():
        my_version = 1
        current_app.logger.info(SSKUpgrader.UPGRADING_MESG.format(my_version))

        SSKUpgrader._to_skip.append(2)
        SSKUpgrader._to_skip.append(3)
        SSKUpgrader._to_skip.append(4)
        SSKUpgrader._to_skip.append(5)
        SSKUpgrader._to_skip.append(6)
        SSKUpgrader._to_skip.append(8)

        set_ssk_version(my_version)

    @staticmethod
    def ver2():
        my_version = 2
        current_app.logger.info(SSKUpgrader.UPGRADING_MESG.format(my_version))

        my_db_version = get_version()

        if my_db_version.ssk_version < my_version and my_version not in SSKUpgrader._to_skip:
            my_engine = create_engine(current_app.config['SQLALCHEMY_BINDS']['logdb'])
            my_engine.execute('alter table access add column response_time INTEGER')

        set_ssk_version(my_version)

    @staticmethod
    def ver3():
        my_version = 3
        current_app.logger.info(SSKUpgrader.UPGRADING_MESG.format(my_version))

        my_db_version = get_version()

        if my_db_version.ssk_version < my_version and my_version not in SSKUpgrader._to_skip:
            get_db().engine.execute('alter table notes add column summary VARCHAR(100)')
            get_db().engine.execute('alter table notes add column category VARCHAR(100)')

        set_ssk_version(my_version)

    @staticmethod
    def ver4():
        my_version = 4
        current_app.logger.info(SSKUpgrader.UPGRADING_MESG.format(my_version))

        my_db_version = get_version()

        if my_db_version.ssk_version < my_version and my_version not in SSKUpgrader._to_skip:
            get_db().engine.execute('alter table notes add column meta VARCHAR(100)')

        set_ssk_version(my_version)

    @staticmethod
    def ver5():
        my_version = 5
        current_app.logger.info(SSKUpgrader.UPGRADING_MESG.format(my_version))

        my_db_version = get_version()

        if my_db_version.ssk_version < my_version and my_version not in SSKUpgrader._to_skip:
            # noinspection PyUnresolvedReferences
            from .models.stats import Stats
            get_db().create_all()

        set_ssk_version(my_version)

    @staticmethod
    def ver6():
        my_version = 6
        current_app.logger.info(SSKUpgrader.UPGRADING_MESG.format(my_version))

        my_db_version = get_version()

        if my_db_version.ssk_version < my_version and my_version not in SSKUpgrader._to_skip:
            # noinspection PyUnresolvedReferences
            from .models.db_stats import DBStats
            get_db().create_all()

        set_ssk_version(my_version)

    @staticmethod
    def ver7():
        my_version = 7
        current_app.logger.info(SSKUpgrader.UPGRADING_MESG.format(my_version))

        my_db_version = get_version()

        if my_db_version.ssk_version < my_version and my_version not in SSKUpgrader._to_skip:
            my_sched_on_setting = Setting()
            my_sched_on_setting.system = True
            my_sched_on_setting.key = "SCHED_ON"
            my_sched_on_setting.type = SettingParser.BOOL_TYPE
            my_sched_on_setting.value = current_app.config["SCHED_ON"]

            my_admin_email = current_app.config["ADMIN_EMAIL"]
            from ssk.models.user import User
            my_admin_user = User.get_by_email(my_admin_email)
            my_admin_user.user_settings.append(my_sched_on_setting)
            get_db().session.commit()

        set_ssk_version(my_version)

    @staticmethod
    def ver8():
        my_version = 8
        current_app.logger.info(SSKUpgrader.UPGRADING_MESG.format(my_version))

        my_db_version = get_version()

        if my_db_version.ssk_version < my_version and my_version not in SSKUpgrader._to_skip:
            # noinspection PyUnresolvedReferences
            from .models.user import Subs, UserSubs, Transaction
            get_db().create_all()

            my_pro_user = User()
            my_pro_user.username = current_app.config['USER_PRO_NAME']
            my_pro_user.email = current_app.config['USER_PRO_EMAIL']
            my_pro_user.first_name = "John"
            my_pro_user.last_name = "Pro"
            my_pro_user.email_confirmed_at = datetime.now()
            from ssk.db import truncate_password
            my_pro_user.password = user_manager.hash_password(truncate_password(current_app.config['USER_PRO_PASS']))

            my_admin_user = User.get_by_email(current_app.config["ADMIN_EMAIL"])
            my_free_user = User.get_by_email(current_app.config["USER_PRO_EMAIL"])

            my_free_subs = Subs()
            my_free_subs.name = current_app.config['FREE_SUB']
            my_free_subs.users.append(my_free_user)

            my_pro_subs = Subs()
            my_pro_subs.name = current_app.config['PRO_SUB']
            my_pro_subs.users.append(my_pro_user)
            my_pro_subs.users.append(my_admin_user)

            func_db.session.add(my_admin_user)
            func_db.session.add(my_free_user)
            func_db.session.add(my_pro_user)
            func_db.session.commit()

        set_ssk_version(my_version)

    @staticmethod
    def get_upgrade_functions():
        my_retval = [SSKUpgrader.ver1, SSKUpgrader.ver2, SSKUpgrader.ver3, SSKUpgrader.ver4, SSKUpgrader.ver5,
                     SSKUpgrader.ver6, SSKUpgrader.ver7, SSKUpgrader.ver8]

        return my_retval



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask_login import current_user

from ...db import get_db
from ...models.audit import Audit


class AbstractCmd:
    ANON = 1
    USER = 2
    MEMBER = 3
    ROOT = 4

    HELP_STR = '[[ print "? for help" ]]'
    PRINT_CMD = '[[ print "\n'

    _supported_cmd = None
    _name = None
    _ctx = None
    _type = None

    def __init__(self, a_name, a_type=ROOT):
        self._supported_cmd = {}
        self._name = a_name
        self._ctx = a_name
        self._type = a_type

    def get_name(self):
        return self._name

    def reg_cmd(self, a_cmd_names: list, a_cmd):
        for my_tmp_cmd_name in a_cmd_names:
            self._supported_cmd[my_tmp_cmd_name] = a_cmd

    def get_all_cmds(self):
        return self._supported_cmd

    def get_params(self, a_params):
        if len(a_params) > 0:
            return a_params.pop(0)

        return "?"

    def get_cmd(self, a_cmd):
        my_retval = None

        if a_cmd in self._supported_cmd:
            my_retval = self._supported_cmd[a_cmd]

        return my_retval

    def sos(self, a_params: list):
        if len(a_params) == 1 and (a_params[0] == "help" or a_params[0] == "?"):
            return True

        return False

    def has_access(self):
        my_user_class = AbstractCmd.ANON

        if not current_user.is_anonymous:
            my_user_class = AbstractCmd.USER

            if current_user.is_admin():
                my_user_class = AbstractCmd.ROOT

        return self._type <= my_user_class

    def exec(self, a_params: list):
        if self.sos(a_params):
            return True, self.help()

        if self.has_access():
            my_ret_val, my_ret_mesg = self.action(a_params)
        else:
            my_ret_val = False
            my_ret_mesg = "Error: cmd access denied"

            my_audit = Audit()
            my_audit.status = "NOK"
            my_audit.category = "CMD"
            if current_user.is_anonymous:
                my_audit.by_user = "Anonymous"
            else:
                my_audit.by_user = current_user.email

            my_audit.description = "access denied cmd"
            get_db().session.add(my_audit)
            get_db().session.commit()

        if my_ret_mesg == '..':
            my_ret_val = True
            my_ret_mesg = ''

        return my_ret_val, my_ret_mesg

    def action(self, a_params: list):
        # placeholder for the real action
        pass

    def help(self, a_wrapped=True):
        # placeholder for teal help
        pass



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#

from datetime import timedelta

from flask import current_app
from sqlalchemy import and_

from .base_job import BaseJob
from ... import get_db
from ...models.access import Access
from statistics import mean

from ...utils import now


class PageStatJob(BaseJob):
    def __init__(self, an_app, a_args):
        super(PageStatJob, self).__init__(an_app, a_args)

    def work(self):
        from ssk.models.stats import Stats

        with self._app.app_context():
            try:
                my_admin_user = current_app.config["ADMIN_NAME"]
                my_processed_id = Stats.get_last_processed_id()

                my_since = now() - timedelta(days=7)

                my_per_page = get_db().session.query(Access).filter(and_(Access.id > my_processed_id,
                                                                         Access.created > my_since,
                                                                         Access.user != my_admin_user)).all()

                my_tmp_mean = {}

                my_db_session = get_db().session
                for my_tmp_access in my_per_page:
                    my_date = "{}-{}-{}".format(my_tmp_access.created.year, my_tmp_access.created.month, my_tmp_access.created.day)

                    if my_tmp_access.response != "200 OK":
                        my_page = "NOTOK"
                    else:
                        my_split = my_tmp_access.path.split("/")
                        if len(my_split) > 0:
                            my_page = my_split[1]

                    if not my_page.startswith("?"):
                        if my_date not in my_tmp_mean.keys():
                            my_tmp_mean[my_date] = {}

                        if my_page not in my_tmp_mean[my_date].keys():
                            my_tmp_mean[my_date][my_page] = []

                        my_stats_db = get_db().session.query(Stats).filter(and_(Stats.day == my_date, Stats.path == my_page)).first()
                        if my_stats_db is None:
                            my_stats_db = Stats()
                            my_stats_db.day = my_date
                            my_stats_db.path = my_page
                            my_stats_db.hits = 0
                            my_stats_db.last_processed = 0
                            my_stats_db.mean_response_time = 0
                            my_stats_db.max_response_time = 0

                        if my_stats_db.last_processed < my_tmp_access.id:
                            my_tmp_mean[my_date][my_page].append(my_tmp_access.response_time)

                            my_stats_db.last_processed = my_tmp_access.id
                            if my_tmp_access.response_time > my_stats_db.max_response_time:
                                my_stats_db.max_response_time = my_tmp_access.response_time

                            my_stats_db.mean_response_time = mean(my_tmp_mean[my_date][my_page])
                            my_stats_db.hits += 1

                        my_db_session.add(my_stats_db)

                my_db_session.commit()

            except Exception as problem:
                self.write_to_log("Page Stat Failed {}".format(problem))



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from .abstract_cmd import AbstractCmd
from .health_cmd import HealthCmd
from .job_cmd import JobCmd
from .page_stat_cmd import PageStatCmd
from .db_stat_cmd import DbStatCmd
from .db_cleanup_cmd import DbCleanupCmd
from .tail_cmd import TailCmd
from .config_cmd import ConfigCmd
from .api_cmd import ApiCmd
from .user_cmd import UserCmd
from .group_cmd import GroupCmd
from .change_pass_cmd import ChangePasswdCmd
from .mail_cmd import MailCmd
from .validate_cmd import ValidateUserCmd
from flask_login import current_user

from ...db import get_db
from ...models.audit import Audit


class AdminCmd(AbstractCmd):
    def __init__(self, a_name="admin"):
        super().__init__(a_name)
        self.reg_cmd(["a", "api"], ApiCmd())
        self.reg_cmd(["c", "conf"], ConfigCmd())
        self.reg_cmd(["g", "group"], GroupCmd())
        self.reg_cmd(["h", "health"], HealthCmd())
        self.reg_cmd(["j", "jobs"], JobCmd())
        self.reg_cmd(["m", "mai;"], MailCmd())
        self.reg_cmd(["p", "passwd"], ChangePasswdCmd())
        self.reg_cmd(["s", "stats"], PageStatCmd())
        self.reg_cmd(["dbst", "dbstats"], DbStatCmd())
        self.reg_cmd(["dbcl", "dbcleanup"], DbCleanupCmd())
        self.reg_cmd(["t", "tail"], TailCmd())
        self.reg_cmd(["u", "user"], UserCmd())
        self.reg_cmd(["v", "verify"], ValidateUserCmd())

    def action(self, a_params):
        my_ret_mesg = self.help()

        my_cmd_name = self.get_params(a_params)
        my_ret_val = (my_cmd_name == "?")
        my_cmd = self.get_cmd(my_cmd_name)
        if my_cmd is not None:
            my_audit = Audit()
            my_params_str = " ".join(a_params)
            my_ret_val, my_ret_mesg = my_cmd.exec(a_params)

            my_audit.category = "CMD"
            my_audit.by_user = current_user.email
            my_audit.description = "{} {}: {}".format(my_cmd_name, my_params_str, my_ret_val)

            if my_ret_val:
                my_audit.status = "OK"
            else:
                my_audit.status = "NOK"
            get_db().session.add(my_audit)
            get_db().session.commit()

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        my_names = []
        my_mesg = '[[ print "---\n'
        for my_tmp_cmd in self.get_all_cmds().values():
            my_name = my_tmp_cmd.get_name()
            if my_name not in my_names:
                my_mesg += "{}\n".format(my_name)
                my_names.append(my_name)

        my_mesg += '" ]]'

        return my_mesg



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app

from ssk.logic.cmd.abstract_cmd import AbstractCmd
from ssk.logic.jobs.db_stat_job import DbStatJob


class DbStatCmd(AbstractCmd):
    def __init__(self):
        super().__init__("dbstat")

    def action(self, a_params: list):
        from ssk.globals.cmd_processor import CmdProcessor

        my_retval = None

        if len(a_params) == 0:
            my_task = DbStatJob(current_app, a_args=["dbstat"])

            CmdProcessor.submit_cmd(my_task)
            my_mesg = '[[ print "OK: started job dbstat {}" ]]'.format(my_task.get_task_id())
            my_retval = my_task.get_task_id()
        else:
            my_mesg = '[[ print "Error: action name is missing" ]]'

        return my_retval, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: dbstat\nstarts a dbstat job" ]]'


#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app
from sqlalchemy import desc

from .abstract_cmd import AbstractCmd
from ..jobs.base_job import BaseJob
from ...models.job import Job
from ...utils import get_padding, get_ago
from ...logic.jobs.empty_job import EmptyJob


class JobPrintCmd(AbstractCmd):
    def __init__(self):
        super().__init__("list")

    def action(self, a_param: list):
        my_mesg = '[[ print "\n'

        my_template = "{} {} {} {} {}\n"
        my_tasks = Job.query.order_by(desc(Job.created))
        my_mesg = my_mesg + my_template.format(get_padding("id", 6),
                                               get_padding("task id", 38),
                                               get_padding("name", 50),
                                               get_padding("status", 12),
                                               get_padding("%", 6))

        my_mesg = my_mesg + my_template.format("_" * 6,
                                               "_" * 38,
                                               "_" * 50,
                                               "_" * 12,
                                               "_" * 6)

        for my_tmp_task in my_tasks:
            my_mesg = my_mesg + my_template.format(get_padding(my_tmp_task.id, 6),
                                                   get_padding(my_tmp_task.task_id, 38),
                                                   get_padding(my_tmp_task.name, 50),
                                                   get_padding(my_tmp_task.status, 12),
                                                   get_padding(my_tmp_task.progress, 6))

            my_time = " "
            if my_tmp_task.done is not None:
                my_time = get_ago((my_tmp_task.done - my_tmp_task.started).seconds, a_suffix="")

            my_start_finish = "s: " + str(my_tmp_task.started)[:19] + " " + "f: " + str(my_tmp_task.done)[:19]

            my_mesg = my_mesg + "{} {} {}\n\n".format(" " * 6,
                                                      get_padding(my_time, 38),
                                                      get_padding(my_start_finish, 50)
                                                      )

        my_mesg = my_mesg + '" ]]'

        return True, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: job ls\nprints the list of all jobs in the system" ]]'


class JobRunCmd(AbstractCmd):
    def __init__(self):
        super().__init__("run")

    def action(self, a_params: list):
        from ssk.globals.cmd_processor import CmdProcessor

        my_retval = None

        if len(a_params) == 1:
            my_args = a_params[0]
            my_task = EmptyJob(current_app, a_args=my_args)

            CmdProcessor.submit_cmd(my_task)
            my_mesg = '[[ print "OK: started job {} {}" ]]'.format(my_args, my_task.get_task_id())
            my_retval = my_task.get_task_id()
        else:
            my_mesg = '[[ print "Error: action name is missing" ]]'

        return my_retval, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: job start <action name>\nstarts a job" ]]'


class JobKillCmd(AbstractCmd):
    def __init__(self):
        super().__init__("kill")

    def action(self, a_params: list):
        from ssk.lg import get_logic

        my_retval = False

        if len(a_params) == 1:
            my_task_id = a_params.pop(0)

            if my_task_id == "all_a":
                my_retval, my_mesg = get_logic().get_job_mgr().stop_all(a_status=BaseJob.IN_PROG_STATUS)
            elif my_task_id == "all_q":
                my_retval, my_mesg = get_logic().get_job_mgr().stop_all(a_status=BaseJob.QUEUED)
            else:
                my_retval, my_mesg = get_logic().get_job_mgr().stop_job(my_task_id)
        else:
            my_mesg = '[[ print "Error: job id is missing" ]]'

        return my_retval, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: job kill <action name>\nstops a job" ]]'


class JobDelCmd(AbstractCmd):
    def __init__(self):
        super().__init__("delete")

    def action(self, a_params: list):
        from ssk.lg import get_logic

        my_retval = False

        if len(a_params) == 1:
            my_task_id = a_params.pop(0)

            if my_task_id == "all_s":
                my_retval, my_mesg = get_logic().get_job_mgr().del_all(a_status=BaseJob.STOPPED_STATUS)
            elif my_task_id == "all_d":
                my_retval, my_mesg = get_logic().get_job_mgr().del_all(a_status=BaseJob.DONE_STATUS)
            else:
                my_retval, my_mesg = get_logic().get_job_mgr().delete_job(my_task_id)
        else:
            my_mesg = '[[ print "Error: job id is missing" ]]'

        return my_retval, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: job del <jobid>\ndeletes a job" ]]'


class JobCmd(AbstractCmd):
    def __init__(self):
        super().__init__("jobs")
        self.reg_cmd(["l", "list"], JobPrintCmd())
        self.reg_cmd(["r", "run"], JobRunCmd())
        self.reg_cmd(["k", "kill"], JobKillCmd())
        self.reg_cmd(["d", "del"], JobDelCmd())

    def action(self, a_params: list):
        my_result = False
        my_mesg = self.help()

        my_cmd_name = self.get_params(a_params)
        my_cmd = self.get_cmd(my_cmd_name)
        if my_cmd is not None:
            my_result, my_mesg = my_cmd.exec(a_params)

        return my_result, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: job {list | run | kill | delete}" ]]'



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app
from ssk.db import get_version, set_version


class DbUpgrader:
    @staticmethod
    def ver1():
        my_version = 1
        current_app.logger.info("upgrading to {}".format(my_version))

        my_db_version = get_version()

        if my_db_version.db_version < my_version:
            set_version(my_version)

    @staticmethod
    def get_upgrade_functions():
        my_retval = [DbUpgrader.ver1]

        return my_retval



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


import uuid

from flask_sqlalchemy import SQLAlchemy
from flask import current_app, g
from datetime import datetime

from sqlalchemy_utils import database_exists, create_database
from ssk.globals.setting_parser import SettingParser

func_db = SQLAlchemy()


def truncate_password(password):
    """Truncate password to 72 bytes for bcrypt compatibility.
    
    Bcrypt has a 72-byte limit. This function truncates passwords to 72 bytes
    to avoid ValueError: password cannot be longer than 72 bytes.
    
    Version: 2025-01-12T14:42:00Z
    """
    if isinstance(password, str):
        password_bytes = password.encode('utf-8')
        if len(password_bytes) > 72:
            return password_bytes[:72].decode('utf-8', errors='ignore')
    return password


def get_db():
    if 'db' not in g:
        g.db = func_db

    return g.db


def db_version_check():
    from .models.audit import Audit
    current_app.logger.info("DB Version Check {}".format(current_app.config["DB_HOST"]))

    if not database_exists(current_app.config["SQLALCHEMY_DATABASE_URI"]):
        current_app.logger.info("Func Database Does Not Exist")

        current_app.logger.info("Creating Func DB")
        create_database(current_app.config["SQLALCHEMY_DATABASE_URI"])

        current_app.logger.info("Creating Log DB")
        create_database(current_app.config["SQLALCHEMY_BINDS"]['logdb'])

        current_app.logger.info("Creating Tables")
        db_create()
        db_init()
        current_app.logger.info("Database {} Created".format(current_app.config["DB_HOST"]))
    else:
        current_app.logger.info("Database Exists")

    my_req_db_ver = int(current_app.config["DB_MODEL_VERSION"])
    my_req_ssk_ver = int(current_app.config["SSK_DB_MODEL_VERSION"])

    my_db_version = None
    try:
        my_db_version = get_version()
    except Exception as a_problem:
        current_app.logger.info("DB Schema still Does Not Exist {}".format(a_problem))

    if my_db_version is None:
        current_app.logger.info("DB Version Not Found")
        current_app.logger.info("Creating Tables")
        db_clean()
        db_create()
        db_init()
        current_app.logger.info("Database {} Created".format(current_app.config["DB_HOST"]))

        my_db_version = get_version()

    current_app.logger.info("DB Ver:{} vs {} SSK DB Ver {} vs {}".format(my_db_version.db_version,
                                                                         my_req_db_ver,
                                                                         my_db_version.ssk_version,
                                                                         my_req_ssk_ver))

    my_stop = False
    if my_db_version.ssk_version < my_req_ssk_ver:
        current_app.logger.info("SSK DB {} < {}".format(my_db_version.ssk_version,
                                                        my_req_ssk_ver))
        upgrade_ssk_db(my_db_version.ssk_version, my_req_ssk_ver)

        my_desc = "SSK DB {} < {}: Upgrade".format(my_db_version.ssk_version, my_req_ssk_ver)
    elif my_db_version.ssk_version > my_req_ssk_ver:
        my_stop = True
        current_app.logger.error("SSK DB Version is ahead of Required Version")
        my_desc = "SSK DB {} > {}: Stop".format(my_db_version.ssk_version, my_req_ssk_ver)
    else:
        current_app.logger.info("SSK DB {} == {}".format(my_db_version.ssk_version,
                                                         my_req_ssk_ver))

        my_desc = "SSK DB {} == {}: No Action Needed".format(my_db_version.ssk_version,
                                                             my_req_ssk_ver)

    if not my_stop:
        if my_db_version.db_version < my_req_db_ver:
            current_app.logger.info("IMP DB {} < {} Upgrading".format(my_db_version.db_version,
                                                                      my_req_db_ver))

            upgrade_imp_db(my_db_version.db_version, my_req_db_ver)
            my_desc = "IMP DB {} < {}: Upgrade".format(my_db_version.db_version,
                                                       my_req_db_ver)

        elif my_db_version.db_version > my_req_db_ver:
            my_stop = True
            current_app.logger.error("IMP DB Version is ahead of Required Version")
            my_desc = "IMP DB {} > {}: Stop".format(my_db_version.db_version,
                                                    my_req_db_ver)
        else:
            current_app.logger.info("IMP DB no upgrade needed")
            my_desc = "IMP DB {} == {}: No Action".format(my_db_version.db_version,
                                                          my_req_db_ver)

    my_audit = Audit()
    my_audit.by_user = "SYSTEM"
    my_audit.category = "START"
    my_audit.description = my_desc
    if my_stop:
        my_audit.status = "NOK"
    else:
        my_audit.status = "OK"
    get_db().session.add(my_audit)
    get_db().session.commit()

    if my_stop:
        raise (SystemExit())


def get_version():
    from .models.version import Version

    return get_db().session.query(Version).order_by(Version.id.desc()).first()


def set_version(a_version):
    from sqlalchemy import text
    from .models.version import Version

    my_db_version = get_db().session.query(Version).order_by(Version.id.desc()).first()
    if my_db_version.db_version < a_version:
        my_sql = 'insert into version (db_version, ssk_version) values({}, {})'.format(a_version,
                                                                                       my_db_version.ssk_version)
        with get_db().engine.connect() as my_connection:
            my_connection.execute(text(my_sql))
            my_connection.commit()


def set_ssk_version(a_version):
    from sqlalchemy import text
    from .models.version import Version

    my_db_version = get_db().session.query(Version).order_by(Version.id.desc()).first()
    if my_db_version.ssk_version < a_version:
        my_sql = 'insert into version (db_version, ssk_version) values({}, {})'.format(my_db_version.db_version,
                                                                                       a_version)
        with get_db().engine.connect() as my_connection:
            my_connection.execute(text(my_sql))
            my_connection.commit()


def upgrade_imp_db(a_from, a_to):
    my_functions = current_app.db_upgrader.get_upgrade_functions()

    for my_tmp_ver in range(a_from, a_to + 1):
        current_app.logger.info("Upgrading IMP DB to {}".format(my_tmp_ver + 1))

        try:
            my_idx = my_tmp_ver-1
            if my_idx >= 0:
                my_functions[my_idx]()
        except Exception as an_e:
            current_app.logger.error("IMP DB Upgrade failed {}".format(str(an_e)))


def upgrade_ssk_db(a_from, a_to):
    from ssk.ssk_upgrader import SSKUpgrader
    my_functions = SSKUpgrader().get_upgrade_functions()

    for my_tmp_ver in range(a_from, a_to + 1):
        current_app.logger.info("Upgrading SSK DB to {}".format(my_tmp_ver + 1))

        try:
            my_idx = my_tmp_ver-1
            if my_idx >= 0:
                my_functions[my_idx]()
        except Exception as an_e:
            current_app.logger.error("SSK DB Upgrade failed {}".format(str(an_e)))


def close_db(e=None):
    current_app.logger.info("Closing Func DB {} e: {}".format(current_app.config["DB_HOST"], e))
    my_db = g.pop('db', None)

    if my_db is not None:
        my_db.session.remove()


def close_logdb(e=None):
    current_app.logger.info("Closing Log DB {} e: {}".format(current_app.config["DB_HOST"], e))
    my_db = g.pop('logdb', None)

    if my_db is not None:
        my_db.session.remove()


# noinspection PyUnresolvedReferences
def db_clean():
    current_app.logger.info("db clean {}".format(current_app.config["DB_HOST"]))

    func_db.session.remove()

    from .models.audit import Audit
    from .models.user import User
    from .models.user import Subs
    from .models.user import UserSubs
    from .models.user import Transaction
    from .models.user import Role
    from .models.user import UserRoles
    from .models.user import UserInvitation
    from .models.version import Version
    from .models.job import Job
    from .models.setting import Setting
    from .models.apikey import ApiKey
    from .models.access import Access
    from .models.status import Status
    from .models.stats import Stats
    from .models.db_stats import DBStats

    func_db.drop_all()
    func_db.session.commit()


# noinspection PyUnresolvedReferences
def db_create():
    current_app.logger.info("db create {}".format(current_app.config["DB_HOST"]))

    from .models.audit import Audit
    from .models.user import User
    from .models.user import Role
    from .models.user import UserRoles
    from .models.user import Subs
    from .models.user import UserSubs
    from .models.user import Transaction
    from .models.user import UserInvitation
    from .models.version import Version
    from .models.job import Job
    from .models.setting import Setting
    from .models.apikey import ApiKey
    from .models.access import Access
    from .models.status import Status
    from .models.stats import Stats
    from .models.db_stats import DBStats

    func_db.create_all()
    func_db.session.commit()


def db_init():
    from .models.version import Version
    from .models.user import User
    from .models.user import Role
    from .models.apikey import ApiKey

    user_manager = current_app.user_manager

    my_db_model_ver = 0
    my_ssk_model_ver = 0
    my_version = Version()
    my_version.db_version = my_db_model_ver
    my_version.ssk_version = my_ssk_model_ver

    current_app.logger.info("db init {} db ver {} ssk db v {}".format(current_app.config["DB_HOST"],
                                                                      my_db_model_ver,
                                                                      my_ssk_model_ver))

    my_admin_api = ApiKey()
    my_admin_api.key = str(uuid.uuid4())

    from .models.setting import Setting
    my_collector_setting = Setting()
    my_collector_setting.system = True
    my_collector_setting.key = "MAX_COLLECTORS"
    my_collector_setting.type = SettingParser.INT_TYPE
    my_collector_setting.value = current_app.config["MAX_COLLECTORS"]

    my_max_active_keys_setting = Setting()
    my_max_active_keys_setting.system = True
    my_max_active_keys_setting.key = "MAX_ACTIVE_KEYS"
    my_max_active_keys_setting.type = SettingParser.INT_TYPE
    my_max_active_keys_setting.value = current_app.config["MAX_ACTIVE_KEYS"]

    my_active_now_setting = Setting()
    my_active_now_setting.system = True
    my_active_now_setting.key = "API_ACTIVE_NOW"
    my_active_now_setting.type = SettingParser.INT_TYPE
    my_active_now_setting.value = current_app.config["API_ACTIVE_NOW"]

    my_website_open_setting = Setting()
    my_website_open_setting.system = True
    my_website_open_setting.key = "WEBSITE_OPEN"
    my_website_open_setting.type = SettingParser.BOOL_TYPE
    my_website_open_setting.value = current_app.config["WEBSITE_OPEN"]

    my_sched_on_setting = Setting()
    my_sched_on_setting.system = True
    my_sched_on_setting.key = "SCHED_ON"
    my_sched_on_setting.type = SettingParser.BOOL_TYPE
    my_sched_on_setting.value = current_app.config["SCHED_ON"]

    my_admin_user = User()
    my_admin_user.username = current_app.config['ADMIN_NAME']
    my_admin_user.email = current_app.config['ADMIN_EMAIL']
    my_admin_user.first_name = "Indiana"
    my_admin_user.last_name = "Jones"
    my_admin_user.email_confirmed_at = datetime.now()
    my_admin_user.password = user_manager.hash_password(truncate_password(current_app.config['ADMIN_PASS']))
    my_admin_user.api_keys.append(my_admin_api)
    my_admin_user.user_settings.append(my_collector_setting)
    my_admin_user.user_settings.append(my_max_active_keys_setting)
    my_admin_user.user_settings.append(my_active_now_setting)
    my_admin_user.user_settings.append(my_website_open_setting)
    my_admin_user.user_settings.append(my_sched_on_setting)

    my_admin_role = Role()
    my_admin_role.name = current_app.config['ADMIN_GROUP_NAME']
    my_admin_role.users.append(my_admin_user)

    my_free_user = User()
    my_free_user.username = current_app.config['USER_FREE_NAME']
    my_free_user.email = current_app.config['USER_FREE_EMAIL']
    my_free_user.first_name = "John"
    my_free_user.last_name = "Free"
    my_free_user.email_confirmed_at = datetime.now()
    my_free_user.password = user_manager.hash_password(truncate_password(current_app.config['USER_FREE_PASS']))

    my_pro_user = User()
    my_pro_user.username = current_app.config['USER_PRO_NAME']
    my_pro_user.email = current_app.config['USER_PRO_EMAIL']
    my_pro_user.first_name = "John"
    my_pro_user.last_name = "Pro"
    my_pro_user.email_confirmed_at = datetime.now()
    my_pro_user.password = user_manager.hash_password(truncate_password(current_app.config['USER_PRO_PASS']))

    my_user_role = Role()
    my_user_role.name = current_app.config['USER_GROUP_NAME']
    my_user_role.users.append(my_free_user)
    my_user_role.users.append(my_pro_user)

    # ver 8
    from .models.user import Subs
    my_free_subs = Subs()
    my_free_subs.name = current_app.config['FREE_SUB']
    my_free_subs.users.append(my_free_user)

    my_pro_subs = Subs()
    my_pro_subs.name = current_app.config['PRO_SUB']
    my_pro_subs.users.append(my_pro_user)
    my_pro_subs.users.append(my_admin_user)

    func_db.session.add(my_version)
    func_db.session.add(my_admin_user)
    func_db.session.add(my_free_user)
    func_db.session.add(my_pro_user)
    func_db.session.commit()



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import render_template
from flask_login import current_user

import ssk


class WebGate:
    PAGE_CLOSED = "ssk/closed.html"
    NOT_FOUND = "ssk/not_found.html"

    @staticmethod
    def is_closed():
        if not ssk.AppSettings().get_setting("WEBSITE_OPEN"):
            if current_user.is_anonymous or not current_user.is_admin():
                return True

        return False

    @staticmethod
    def render_closed():
        return render_template(WebGate.PAGE_CLOSED, admin_group_name=ssk.SSK_ADMIN_GROUP)

    @staticmethod
    def render_notfound():
        return render_template(WebGate.NOT_FOUND, admin_group_name=ssk.SSK_ADMIN_GROUP)



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from sqlalchemy import desc

from .abstract_cmd import AbstractCmd
from ...models.apikey import ApiKey
from ...globals.api_gate import ApiGate
from ...utils import get_padding
from ...models.user import User
from ...db import get_db


class ApiPrintCmd(AbstractCmd):
    def __init__(self):
        super().__init__("list")

    def action(self, a_param: list):
        my_ret_val = True
        my_mesg = '[[ print "\n'

        my_status = "OPEN"
        if not ApiGate.is_open():
            my_status = "CLOSED"

        my_template = "{} {} {} {} {}\n"
        my_apis = ApiKey.query.order_by(desc(ApiKey.created)).all()
        my_mesg = my_mesg + "API gate:     {}\n#active keys: {} of {} total {}\n\n".format(my_status,
                                                                                           len(ApiGate.active_now()),
                                                                                           ApiGate.num_total_keys(),
                                                                                           len(my_apis))

        my_mesg = my_mesg + my_template.format(get_padding("db id", 6),
                                               get_padding("owner", 30),
                                               get_padding("key", 40),
                                               get_padding("active", 7),
                                               get_padding("created", 20))
        my_mesg = my_mesg + my_template.format("_" * 6,
                                               "_" * 30,
                                               "_" * 40,
                                               "_" * 7,
                                               "_" * 20)

        for my_tmp_api in my_apis:
            my_mesg = my_mesg + my_template.format(get_padding(my_tmp_api.id, 6),
                                                   get_padding(my_tmp_api.owner.email, 30),
                                                   get_padding(my_tmp_api.key, 40),
                                                   get_padding(my_tmp_api.active, 7),
                                                   get_padding(my_tmp_api.created, 20))

        my_mesg = my_mesg + '" ]]'

        return my_ret_val, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: api ls\nprints the list of all api keys" ]]'


class ApiOpenCmd(AbstractCmd):
    def __init__(self):
        super().__init__("open")

    def action(self, a_param: list):
        ApiGate.set_open(True)

        return True, '[[ print "OK: api is now open" ]]'

    def help(self, a_wrapped=True):
        return '[[ print "Usage: api open\nopens API Gate" ]]'


class ApiCloseCmd(AbstractCmd):
    def __init__(self):
        super().__init__("close")

    def action(self, a_param: list):
        ApiGate.set_open(False)

        return True, '[[ print "OK: api closed" ]]'

    def help(self, a_wrapped=True):
        return '[[ print "Usage: api close\ncloses API Gate" ]]'


class ApiAddCmd(AbstractCmd):
    def __init__(self):
        super().__init__("add")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = self.help()
        if len(a_params) == 1:
            my_user = User.get_by_email(a_params[0])

            if my_user is not None:
                ApiGate.add_api_key(a_params[0])

                my_ret_val = True
                my_ret_mesg = '[[ print "OK: api key added to {}" ]]'.format(a_params[0])
            else:
                my_ret_mesg = '[[ print "Error: User {} not found" ]]'.format(a_params[0])

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: api add <email>\nadds an active api key for a user" ]]'


class ApiDisableCmd(AbstractCmd):
    def __init__(self):
        super().__init__("off")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = self.help()

        if len(a_params) == 1:
            my_key = a_params[0]
            my_api_key = ApiKey.get_by_key(my_key)

            if my_api_key is not None:
                my_api_key.active = False
                get_db().session.add(my_api_key)
                get_db().session.commit()
                ApiGate.load()

                my_ret_val = True
                my_ret_mesg = '[[ print "OK: api key {} disabled" ]]'.format(a_params[0])
            else:
                my_ret_mesg = '[[ print "Error: key {} not found" ]]'.format(a_params[0])

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: api off <key>\ndisables an api key" ]]'


class ApiEnableCmd(AbstractCmd):
    def __init__(self):
        super().__init__("on")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = self.help()

        if len(a_params) == 1:
            my_key = a_params[0]
            my_api_key = ApiKey.get_by_key(my_key)

            if my_api_key is not None:
                my_api_key.active = True
                get_db().session.add(my_api_key)
                get_db().session.commit()
                ApiGate.load()

                my_ret_val = True
                my_ret_mesg = '[[ print "OK: api key {} enabled" ]]'.format(a_params[0])
            else:
                my_ret_mesg = '[[ print "Error: key {} not found" ]]'.format(a_params[0])

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: api on <key>\nenables an api key" ]]'


class ApiCmd(AbstractCmd):
    def __init__(self):
        super().__init__("api")
        self.reg_cmd(["l", "list"], ApiPrintCmd())
        self.reg_cmd(["o", "open"], ApiOpenCmd())
        self.reg_cmd(["c", "close"], ApiCloseCmd())
        self.reg_cmd(["a", "add"], ApiAddCmd())
        self.reg_cmd(["d", "disable"], ApiDisableCmd())
        self.reg_cmd(["e", "enable"], ApiEnableCmd())

    def action(self, a_params: list):
        my_result = False
        my_mesg = self.help()

        my_cmd_name = self.get_params(a_params)
        my_cmd = self.get_cmd(my_cmd_name)
        if my_cmd is not None:
            my_result, my_mesg = my_cmd.exec(a_params)

        return my_result, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: api {list | open | close | add | enable | disable }" ]]'



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from time import time
import json
import uuid

from ..globals.app_settings import AppSettings
from ..models.apikey import ApiKey
from ..utils import get_safe_string


class ApiGate:
    DONE = json.dumps({"code": 0, "mesg": "ok"})
    INVALID_KEY = json.dumps({"code": 10, "mesg": "invalid api key"})
    TOO_BUSY = json.dumps({"code": 20, "mesg": "maximum active sessions exceeded"})
    CLOSED = json.dumps({"code": 30, "mesg": "API is temporarily disabled"})

    __allowed = None
    __active = None
    __timeout = None
    __open = False

    @staticmethod
    def init():
        my_app_settings = AppSettings()

        ApiGate.__timeout = my_app_settings.get_setting("API_ACTIVE_NOW")
        ApiGate.__active = {}
        ApiGate.__allowed = {}

    @staticmethod
    def load():
        my_keys = ApiKey.get_all_active()

        ApiGate.__allowed = {}
        for my_tmp_key in my_keys:
            ApiGate.__allowed[my_tmp_key.key] = "key driver {}".format(my_tmp_key.id)

    @staticmethod
    def get_new_key():
        my_ret_val = str(uuid.uuid4())

        return my_ret_val

    @staticmethod
    def add_api_key(a_user_email):
        my_ret_val = None
        from ssk.models.user import User

        my_user = User.get_by_email(a_user_email)
        if my_user is not None:
            from ssk.db import get_db

            my_api_key = ApiKey()
            my_api_key.owner = my_user
            my_api_key.key = ApiGate.get_new_key()
            get_db().session.add(my_api_key)
            get_db().session.commit()

            ApiGate.load()
            my_ret_val = my_api_key.key

        return my_ret_val

    @staticmethod
    def active_now():
        my_now = time()
        my_threshold = my_now - ApiGate.__timeout

        my_values = ApiGate.__active.values()
        my_active = list(filter(lambda my_key: my_key > my_threshold, my_values))

        return my_active

    @staticmethod
    def is_active_now(a_key):
        my_retval = False

        if ApiGate.__active is not None and a_key in ApiGate.__active.keys():
            my_ts = ApiGate.__active[a_key]
            my_retval = (my_ts > time() - ApiGate.__timeout)

        if my_retval:
            ApiGate.__active[a_key] = time()

        return my_retval

    @staticmethod
    def num_active_keys():
        if ApiGate.__active is None:
            return 0

        return len(ApiGate.active_now())

    @staticmethod
    def num_total_keys():
        if ApiGate.__allowed is None:
            return 0

        return len(ApiGate.__allowed.keys())

    @staticmethod
    def set_open(a_flag):
        ApiGate.__open = a_flag

    @staticmethod
    def is_open():
        return ApiGate.__open

    @staticmethod
    def is_valid(an_app, a_key, ignore_open=False):
        my_ret_val = False

        if ApiGate.is_open() or ignore_open:
            if ApiGate.__allowed is not None:
                my_ret_val = a_key in ApiGate.__allowed.keys()

                if my_ret_val:
                    ApiGate.__active[a_key] = time()

            if not my_ret_val:
                an_app.logger.info("api key rejected %s", get_safe_string(a_key))
            else:
                an_app.logger.debug("api key accepted %s", get_safe_string(a_key))
        else:
            an_app.logger.debug("api is closed")

        return my_ret_val



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


import os

from flask import current_app
from sqlalchemy import asc

from ...models.audit import Audit
from ...utils import get_padding, get_timestamp_str
from .abstract_cmd import AbstractCmd


class TailAuditCmd(AbstractCmd):
    def __init__(self):
        super().__init__("audit")

    def action(self, a_param: list):
        my_mesg = '[[ print "\n'
        my_audit_list = []

        if len(a_param) > 0:
            if a_param[0] == "nok":
                my_audit_list = Audit.query.filter(Audit.status != "OK").order_by(asc(Audit.created)).limit(1000)
        else:
            my_audit_list = Audit.query.order_by(asc(Audit.created)).limit(1000)

        my_template = "{} {} {} {} {}\n"
        my_mesg = my_mesg + my_template.format(get_padding("id", 6),
                                               get_padding("by", 25),
                                               get_padding("category", 10),
                                               get_padding("status", 10),
                                               get_padding("created", 20))

        my_mesg = my_mesg + my_template.format("_" * 6,
                                               "_" * 25,
                                               "_" * 10,
                                               "_" * 10,
                                               "_" * 20)

        for my_tmp_audit in my_audit_list:
            my_description = my_tmp_audit.description
            my_description = my_description.replace("`", "")

            my_mesg = my_mesg + my_template.format(get_padding(my_tmp_audit.id, 6),
                                                   get_padding(my_tmp_audit.by_user, 25),
                                                   get_padding(my_tmp_audit.category, 10),
                                                   get_padding(my_tmp_audit.status, 10),
                                                   get_padding(my_tmp_audit.created, 20))

            my_mesg = my_mesg + "{} {}\n".format(" " * 6, get_padding(my_description, 250))

        my_mesg = my_mesg + '" ]]'

        return True, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: tail audit {nok}\nprints the list of 1000 recent audit entries" ]]'


class TailLogCmd(AbstractCmd):
    def __init__(self):
        super().__init__("log")

    def action(self, a_params: list):
        my_result = False
        my_mesg = self.help()

        my_lines = 100
        if len(a_params) == 1:
            try:
                my_lines = int(a_params.pop(0))
                if my_lines > 10000:
                    my_lines = 0
                    my_mesg = "Error: parameter must be an integer (0-10000)"

            except Exception:
                my_lines = 0
                my_mesg = "Error: parameter must be an integer (0-10000)"

        if my_lines > 0:
            my_mesg = '[[ print "\n'
            my_bufsize = 8192
            my_fname = current_app.config["LOG_FILE"]
            my_fsize = os.stat(my_fname).st_size

            my_iter = 0
            with open(my_fname) as f:
                if my_bufsize > my_fsize:
                    my_bufsize = my_fsize - 1

                data = []

                while True:
                    my_iter += 1
                    my_loc = my_fsize - my_bufsize * my_iter
                    if my_loc < 0:
                        my_loc = my_fsize - my_bufsize

                        f.seek(my_loc)
                        data.extend(f.readlines())
                        if len(data) >= my_lines or f.tell() == 0:
                            my_line = "{}\n".format(''.join(data[-my_lines:]))
                            my_line = my_line.replace("[", "")
                            my_line = my_line.replace("]", "")
                            my_line = my_line.replace("'", "")
                            my_line = my_line.replace("\"", "")
                            my_line = my_line.replace("\u001b", "")

                            my_mesg = my_mesg + my_line
                            break

            my_result = True
            my_mesg = my_mesg + '" ]]'
            my_mesg = '[[ ts "{}" ]]'.format(get_timestamp_str()) + my_mesg

        return my_result, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: tail error [num_of_lines]" ]]'


class TailCmd(AbstractCmd):
    def __init__(self):
        super().__init__("tail")
        self.reg_cmd(["l", "log"], TailLogCmd())
        self.reg_cmd(["a", "audit"], TailAuditCmd())

    def action(self, a_params: list):
        my_result = False
        my_mesg = self.help()

        my_cmd_name = self.get_params(a_params)
        my_cmd = self.get_cmd(my_cmd_name)
        if my_cmd is not None:
            my_result, my_mesg = my_cmd.exec(a_params)

        return my_result, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: tail {log | audit}" ]]'



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask_wtf import FlaskForm
from wtforms import TextAreaField, SubmitField, StringField, HiddenField
from wtforms.validators import DataRequired, Email


class ContactForm(FlaskForm):
    name = StringField('Name', validators=[DataRequired()])
    email = StringField('Email', validators=[DataRequired(), Email()])
    message = TextAreaField("Additional Information")
    submit = SubmitField("Please contact me")
    source = HiddenField("source")



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


import json

from flask import current_app

from .abstract_cmd import AbstractCmd
from ...models.setting import Setting
from ...utils import get_padding
from ...globals.app_settings import AppSettings
from ...globals.app_settings import SettingParser


class ConfigConstsCmd(AbstractCmd):
    def __init__(self):
        super().__init__("consts")

    def action(self, a_params: list):
        my_retval = True
        my_mesg = '[[ print "\n'

        my_template = "{} {}\n"
        my_mesg = my_mesg + my_template.format(get_padding("name", 20),
                                               get_padding("value", 50))

        my_mesg = my_mesg + my_template.format("_" * 20, "_" * 50)

        my_mesg = my_mesg + my_template.format(get_padding("MAIL_USERNAME", 20), current_app.config["MAIL_USERNAME"])
        my_mesg = my_mesg + my_template.format(get_padding("MAIL_PASSWORD", 20), current_app.config["MAIL_PASSWORD"])
        my_mesg = my_mesg + my_template.format(get_padding("MAIL_SUPPRESS_SEND", 20),
                                               current_app.config["MAIL_SUPPRESS_SEND"])

        my_mesg = my_mesg + my_template.format(get_padding("MAX_COLLECTORS", 20), current_app.config["MAX_COLLECTORS"])
        my_mesg = my_mesg + my_template.format(get_padding("MAX_ACTIVE_KEYS", 20),
                                               current_app.config["MAX_ACTIVE_KEYS"])
        my_mesg = my_mesg + my_template.format(get_padding("API_ACTIVE_NOW", 20), current_app.config["API_ACTIVE_NOW"])

        my_mesg = my_mesg + '" ]]'

        return my_retval, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: conf consts\nprints all constants" ]]'


class ConfigPrintCmd(AbstractCmd):
    def __init__(self):
        super().__init__("list")

    def action(self, a_params: list):
        my_retval = True
        my_mesg = '[[ print "\n'

        my_template = "{} {} {} {} {} {}\n"
        my_settings = Setting.query.all()
        my_mesg = my_mesg + my_template.format(get_padding("key", 20),
                                               get_padding("owner", 30),
                                               get_padding("global", 7),
                                               get_padding("type", 7),
                                               get_padding("updated", 20),
                                               get_padding("created", 20)
                                               )

        my_mesg = my_mesg + my_template.format("_" * 20,
                                               "_" * 30,
                                               "_" * 7,
                                               "_" * 7,
                                               "_" * 20,
                                               "_" * 20)

        for my_tmp_setting in my_settings:
            my_mesg = my_mesg + my_template.format(get_padding(my_tmp_setting.key, 20),
                                                   get_padding(my_tmp_setting.owner.email, 30),
                                                   get_padding(my_tmp_setting.system, 7),
                                                   get_padding(my_tmp_setting.type, 7),
                                                   get_padding(my_tmp_setting.updated, 20),
                                                   get_padding(my_tmp_setting.created, 20))

            my_value = my_tmp_setting.value
            if my_tmp_setting.type == SettingParser.JSON_TYPE:
                my_json = json.loads(my_value)
                for my_tmp_key in my_json.keys():
                    my_mesg = my_mesg + "{} {} {} : {}\n".format(" " * 20, " " * 30, my_tmp_key, my_json[my_tmp_key])
            else:
                my_mesg = my_mesg + "{} {} {}\n".format(" " * 20, " " * 30, get_padding(my_value, 100))

            my_mesg = my_mesg + "\n"

        my_mesg = my_mesg + '" ]]'

        return my_retval, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: conf ls\nprints all settings" ]]'


class ConfigGlobalCmd(AbstractCmd):
    def __init__(self):
        super().__init__("global")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = AbstractCmd.HELP_STR

        if len(a_params) == 2:
            my_key = a_params[0]
            my_on = a_params[1]

            my_global = False
            if my_on == "on":
                my_global = True

            my_ret_val, my_ret_mesg = AppSettings().set_global(my_key, my_global)

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: conf glb <key> [on | off]\nmakes a setting global" ]]'


class ConfigSetCmd(AbstractCmd):
    def __init__(self):
        super().__init__("set")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = AbstractCmd.HELP_STR

        if len(a_params) == 4:
            my_user = a_params[0]
            my_type = a_params[1]
            my_key = a_params[2]
            my_value = a_params[3]

            my_ret_val, my_ret_mesg = AppSettings().set_setting(my_user, my_key, my_type, my_value)

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: conf set <email> <type> <key> <value>\nsets or adds a setting for a given user" ]]'


class ConfigGetCmd(AbstractCmd):
    def __init__(self):
        super().__init__("get")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = AbstractCmd.HELP_STR

        if len(a_params) == 2:
            my_key = a_params[0]
            my_user = a_params[1]

            my_result = AppSettings().get_setting(my_key, my_user)
            my_ret_val = True
            my_ret_mesg = '[[ print "{} for {} = {}" ]]'.format(my_key, my_user, my_result)

        if len(a_params) == 1:
            my_key = a_params[0]

            my_result = AppSettings().get_setting(my_key)

            my_ret_val = True
            my_ret_mesg = '[[ print "{} = {}" ]]'.format(my_key, my_result)

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: conf get <key> <email>\ngets a setting for a given user" ]]'


class ConfigDelCmd(AbstractCmd):
    def __init__(self):
        super().__init__("delete")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = AbstractCmd.HELP_STR

        if len(a_params) == 2:
            my_user = a_params[0]
            my_key = a_params[1]

            my_ret_val, my_ret_mesg = AppSettings().del_setting(my_user, my_key)

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: conf del <email> <key>\ndeletes a setting for a given user" ]]'


class ConfigCmd(AbstractCmd):
    def __init__(self):
        super().__init__("config")
        self.reg_cmd(["l", "list"], ConfigPrintCmd())
        self.reg_cmd(["c", "consts"], ConfigConstsCmd())
        self.reg_cmd(["s", "set"], ConfigSetCmd())
        self.reg_cmd(["g", "get"], ConfigGetCmd())
        self.reg_cmd(["d", "delete"], ConfigDelCmd())
        self.reg_cmd(["gl", "global"], ConfigGlobalCmd())

    def action(self, a_params: list):
        my_result = False
        my_mesg = self.help()

        my_cmd_name = self.get_params(a_params)
        my_cmd = self.get_cmd(my_cmd_name)
        if my_cmd is not None:
            my_result, my_mesg = my_cmd.exec(a_params)

        return my_result, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: conf {list | consts | set | get | delete | global }" ]]'



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from sqlalchemy import desc

from .abstract_cmd import AbstractCmd
from ...models.user import User, Role
from ...utils import get_padding
from ...db import get_db


class GroupPrintCmd(AbstractCmd):
    def __init__(self):
        super().__init__("list")

    def action(self, a_param: list):
        my_mesg = '[[ print "\n'

        my_template = "{} {} {} {}\n"
        my_roles = Role.query.order_by(desc(Role.created)).all()
        my_mesg = my_mesg + my_template.format(get_padding("id", 6),
                                               get_padding("name", 10),
                                               get_padding("members", 30),
                                               get_padding("created", 20))

        my_mesg = my_mesg + my_template.format("_" * 6,
                                               "_" * 10,
                                               "_" * 30,
                                               "_" * 20)

        for my_tmp_role in my_roles:
            my_members = my_tmp_role.users
            my_mesg = my_mesg + my_template.format(get_padding(my_tmp_role.id, 6),
                                                   get_padding(my_tmp_role.name, 10),
                                                   get_padding(len(my_members), 30),
                                                   get_padding(my_tmp_role.created, 20))

            for my_tmp_member in my_members:
                my_mesg = my_mesg + my_template.format(" " * 6,
                                                       " " * 10,
                                                       get_padding(my_tmp_member.email, 30),
                                                       "  " * 20)

            my_mesg = my_mesg + my_template.format(" " * 6,
                                                   " " * 10,
                                                   " " * 30,
                                                   " " * 20)

        my_mesg = my_mesg + '" ]]'

        return True, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: group ls\nprints list of all groups" ]]'


class GroupAddCmd(AbstractCmd):
    def __init__(self):
        super().__init__("add")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = AbstractCmd.HELP_STR

        if len(a_params) == 2:
            my_user_name = a_params[0]
            my_role_name = a_params[1]

            my_role = get_db().session.query(Role).filter(Role.name == my_role_name).first()
            if my_role is not None:
                my_user = get_db().session.query(User).filter(User.email == my_user_name).first()
                if my_user is not None:
                    my_user.roles.append(my_role)

                    my_ret_val = True
                    my_ret_mesg = '[[ print "OK: {} added to {}" ]]'.format(my_role_name, my_role.name)
                else:
                    my_ret_mesg = '[[ print "Error: user {} not found" ]]'.format(my_user_name)
            else:
                my_ret_mesg = '[[ print "Error: group \'{}\' not found" ]]'.format(my_role_name)

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: group add <email> <group_name> \nadd user to a group" ]]'


class GroupDelCmd(AbstractCmd):
    def __init__(self):
        super().__init__("delete")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = AbstractCmd.HELP_STR

        if len(a_params) == 2:
            my_user_name = a_params[0]
            my_role_name = a_params[1]

            my_role = get_db().session.query(Role).filter(Role.name == my_role_name).first()
            if my_role is not None:
                my_user = get_db().session.query(User).filter(User.email == my_user_name).first()
                if my_user is not None:
                    if my_user.has_role(my_role.name):
                        my_user.roles.remove(my_role)

                        my_ret_val = True
                        my_ret_mesg = '[[ print "OK: {} removed from {}" ]]'.format(my_user_name, my_role_name)
                    else:
                        my_ret_mesg = '[[ print "Error: {} is not {}" ]]'.format(my_user_name, my_role_name)
                else:
                    my_ret_mesg = '[[ print "Error: user {} not found" ]]'.format(my_user_name)
            else:
                my_ret_mesg = '[[ print "Error: group {} not found" ]]'.format(my_role_name)

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: group del <email> <group_name> \ndelete user from a group" ]]'


class GroupNewCmd(AbstractCmd):
    def __init__(self):
        super().__init__("new")

    def action(self, a_params: list):
        my_ret_val = False
        my_ret_mesg = AbstractCmd.HELP_STR

        if len(a_params) == 1:
            my_role_name = a_params[0]

            my_role = get_db().session.query(Role).filter(Role.name == my_role_name).first()
            if my_role is None:
                my_new_role = Role()
                my_new_role.name = my_role_name

                get_db().session.add(my_new_role)
                get_db().session.commit()

                my_ret_val = True
                my_ret_mesg = '[[ print "OK: group \'{}\' created" ]]'.format(my_role_name)
            else:
                my_ret_mesg = '[[ print "Error: group \'{}\' already exists" ]]'.format(my_role_name)

        return my_ret_val, my_ret_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: group new <group_name> \ncreates new user group" ]]'


class GroupCmd(AbstractCmd):
    def __init__(self):
        super().__init__("groups")
        self.reg_cmd(["l", "list"], GroupPrintCmd())
        self.reg_cmd(["n", "new"], GroupNewCmd())
        self.reg_cmd(["a", "add"], GroupAddCmd())
        self.reg_cmd(["d", "delete"], GroupDelCmd())

    def action(self, a_params: list):
        my_result = False
        my_mesg = self.help()

        my_cmd_name = self.get_params(a_params)
        my_cmd = self.get_cmd(my_cmd_name)
        if my_cmd is not None:
            my_result, my_mesg = my_cmd.exec(a_params)

        return my_result, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: groups {list | new | add | delete}" ]]'



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


import os.path

from flask import (Blueprint, current_app, send_file)
from flask import render_template
from flask_login import current_user
from flask_user import roles_required

from .. import get_db
from ..models.audit import Audit
from ..utils import get_timestamp_str
from ..ssk_consts import SSK_ADMIN_GROUP

from ssk.blueprints.admin_handler import AdminHandler

bp = Blueprint('admin', __name__, url_prefix='/admin')


@bp.route('/', methods=['GET'])
@roles_required(SSK_ADMIN_GROUP)
def index():
    return render_template('ssk/home.html', timestamp=get_timestamp_str())


@bp.route('/log_download', methods=['GET'])
@roles_required(SSK_ADMIN_GROUP)
def log_download():
    my_current = os.getcwd()
    my_log_path = os.path.join(my_current, current_app.config["LOG_FILE"])

    my_status = "OK"
    if not os.path.exists(my_log_path):
        my_status = "NOK"

    my_audit = Audit()
    my_audit.status = my_status
    my_audit.category = "LOG"
    my_audit.by_user = current_user.email
    my_audit.description = "log {}".format(my_log_path)
    get_db().session.add(my_audit)
    get_db().session.commit()

    if my_status == "OK":
        return send_file(my_log_path, as_attachment=True)

    return render_template('ssk/home.html', timestamp=get_timestamp_str())


@bp.route('/system_chart', methods=['GET'])
@roles_required(SSK_ADMIN_GROUP)
def system_chart():
    my_file, my_to_plot = AdminHandler.system_chart()

    return render_template(my_file, system_status=my_to_plot, admin_group_name=SSK_ADMIN_GROUP)


@bp.route('/system_perf', methods=['GET'])
@roles_required(SSK_ADMIN_GROUP)
def system_perf():
    my_file, my_stats = AdminHandler.system_stats()

    return render_template(my_file,
                           system_perf=my_stats["plot"],
                           daily_stats=my_stats["stats"],
                           daily_perf=my_stats["perf"],
                           all_pages=my_stats["all_pages"],
                           all_days=my_stats["all_days"],
                           db_stats=my_stats["db_stats"],
                           db_tables=my_stats["db_tables"],
                           db_days=my_stats["db_days"],
                           admin_group_name=SSK_ADMIN_GROUP)



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import (Blueprint, request, render_template)
from flask_user import roles_required

from ssk.ssk_consts import SSK_ADMIN_GROUP
from ssk.blueprints.cmd_handler import CmdHandler

bp = Blueprint('cmd', __name__, url_prefix='/cmd')


@bp.route('/terminal', methods=['GET'])
@roles_required(SSK_ADMIN_GROUP)
def terminal():
    my_page = CmdHandler.get_terminal()

    return render_template(my_page, admin_group_name=SSK_ADMIN_GROUP)


@bp.route('/cmd', methods=['POST'])
@roles_required(SSK_ADMIN_GROUP)
def cmd():
    return CmdHandler.cmd_post(request)



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from sqlalchemy import desc

from .abstract_cmd import AbstractCmd
from ...models.status import Status
from ...utils import get_padding


class HealthCmd(AbstractCmd):
    def __init__(self):
        super().__init__("health")

    def action(self, a_param: list):
        my_mesg = ''

        if len(a_param) == 0:
            my_mesg = '[[ print "\n'

            my_template = "{} {} {} {} {} {} {} {} {} {} {} {}\n"
            my_status = Status.query.order_by(desc(Status.created)).limit(60)
            my_mesg = my_mesg + my_template.format(get_padding("api", 8),
                                                   get_padding("thr", 4),
                                                   get_padding("now", 4),
                                                   get_padding("max", 4),
                                                   get_padding("tot", 4),
                                                   get_padding("usr", 4),
                                                   get_padding("now", 4),
                                                   get_padding("cnf", 4),
                                                   get_padding("aud", 4),
                                                   get_padding("rt", 4),
                                                   get_padding("mem", 6),
                                                   get_padding("created", 20),
                                                   )

            my_mesg = my_mesg + my_template.format("_" * 8,
                                                   "_" * 4,
                                                   "_" * 4,
                                                   "_" * 4,
                                                   "_" * 4,
                                                   "_" * 4,
                                                   "_" * 4,
                                                   "_" * 4,
                                                   "_" * 4,
                                                   "_" * 4,
                                                   "_" * 6,
                                                   "_" * 20)

            for my_tmp_status in my_status:
                my_mesg = my_mesg + my_template.format(get_padding(my_tmp_status.api_status, 8),
                                                       get_padding(my_tmp_status.api_now_threshold, 4),
                                                       get_padding(my_tmp_status.api_active_now, 4),
                                                       get_padding(my_tmp_status.api_max_active, 4),
                                                       get_padding(my_tmp_status.api_total, 4),
                                                       get_padding(my_tmp_status.users, 4),
                                                       get_padding(my_tmp_status.users_active, 4),
                                                       get_padding(my_tmp_status.conf_keys, 4),
                                                       get_padding(my_tmp_status.audit_cnt, 4),
                                                       get_padding(my_tmp_status.response_time, 4),
                                                       get_padding(my_tmp_status.mem, 6),
                                                       get_padding(my_tmp_status.created, 20))

            my_mesg = my_mesg + '" ]]'
        elif a_param[0] == "h":
            my_mesg = self.help()

        return True, my_mesg

    def help(self, a_wrapped=True):
        return '[[ print "Usage: health\nprints last 60 system health snapshots" ]]'



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask import current_app, render_template

from flask_user import current_user
from flask_mail import Mail, Message

from .app_settings import AppSettings
from ..models.audit import Audit


class EmailMgr:
    _db_session = None

    def __init__(self, a_db_session):
        self._db_session = a_db_session

    def sendit(self, a_message):
        my_email_off = AppSettings().get_setting("IS_OFFLINE")
        if my_email_off is None or my_email_off is False:
            my_mail = Mail(current_app)
            my_mail.send(a_message)

    def send_test(self, a_from, a_to):
        try:
            my_mail_mesg = Message('Test email', sender=a_from, recipients=[a_to])

            my_render_args = {'app_name': current_app.config["USER_APP_NAME"],
                              'app_version': current_app.config["USER_APP_VERSION"], 'user': current_user}

            my_txt_message = render_template('ssk/emails/test.txt', **my_render_args)
            my_html_message = render_template('ssk/emails/test.html', **my_render_args)

            my_mail_mesg.body = my_txt_message
            my_mail_mesg.html = my_html_message

            self.sendit(my_mail_mesg)

            # keeping track of things
            my_audit = Audit()
            my_audit.category = "EMAIL"
            my_audit.by_user = "admin"
            my_audit.status = "OK"
            my_audit.description = "Test email sent from {} to {}".format(a_from, a_to)
        except Exception as ex:
            my_audit = Audit()
            my_audit.category = "EMAIL"
            my_audit.by_user = "admin"
            my_audit.status = "NOK"
            my_audit.description = "Error: Test email sent from {} to {} failed {}".format(a_from, a_to, ex)

        self._db_session.add(my_audit)
        return my_audit


#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from sqlalchemy.orm import relationship

from ..db import func_db, get_db


class Job(func_db.Model):
    __tablename__ = 'job'

    id = func_db.Column(func_db.Integer, primary_key=True)
    task_id = func_db.Column(func_db.String(50), unique=False, nullable=False)
    name = func_db.Column(func_db.String(25), unique=False, nullable=False)
    action = func_db.Column(func_db.String(50), unique=False, nullable=False)
    status = func_db.Column(func_db.String(50), unique=False, nullable=False)
    logfile = func_db.Column(func_db.String(50), unique=False, nullable=True)
    progress = func_db.Column(func_db.Integer)

    user_id = func_db.Column(func_db.Integer(), func_db.ForeignKey('user.id', ondelete='CASCADE'))
    executor = relationship("User", back_populates="jobs")

    created = func_db.Column(func_db.DateTime(timezone=True), server_default=func_db.func.current_timestamp())
    started = func_db.Column(func_db.DateTime(timezone=True))
    done = func_db.Column(func_db.DateTime(timezone=True))

    @staticmethod
    def get_by_key(a_key):
        my_ret_val = get_db().session.query(Job).filter(Job.task_id == a_key).first()

        return my_ret_val

    def __repr__(self):
        return '<Job %r>' % self.task_id



#
# Copyright (c) 2023 Michał Świtała / CodingMinds.io
# SPDX-License-Identifier: MIT
#


from flask_login import current_user
from ssk import AppSettings, get_logic, get_db
from ssk.globals.contacts_mgr import ContactMgr
from ssk.forms.contact_form import ContactForm


class HomeHandler(object):
    PAGE_CLOSED = "ssk/closed.html"
    PAGE_TASKS = "ssk/tasks.html"
    PAGE_NOTES = "ssk/notes.html"
    PAGE_ABOUT = "ssk/about.html"

    @staticmethod
    def tasks():
        my_tasks = []

        if AppSettings().get_setting("LOGIN_OFF"):
            return HomeHandler.PAGE_CLOSED, my_tasks

        if not current_user.is_anonymous:
            my_jobs = get_logic().get_job_mgr().get_all_jobs_by_user(current_user.id)

            for my_temp_task in my_jobs:
                my_tasks.append({"tid": my_temp_task.task_id,
                                 "name": my_temp_task.name,
                                 "details": my_temp_task.started.strftime("%d-%b-%Y %H:%M:%S"),
                                 "status": my_temp_task.status,
                                 "prog": my_temp_task.progress})

        return HomeHandler.PAGE_TASKS, my_tasks

    @staticmethod
    def tasks_action(a_request):
        my_tasks = []

        if AppSettings().get_setting("LOGIN_OFF"):
            return HomeHandler.PAGE_CLOSED, my_tasks

        if a_request.method == 'POST':
            my_job_mgr = get_logic().get_job_mgr()

            my_job_to_stop = a_request.form.get('stop')
            if my_job_to_stop is not None:
                my_job_mgr.stop_job(my_job_to_stop)

            my_job_to_delete = a_request.form.get('delete')
            if my_job_to_delete is not None:
                my_job_mgr.delete_job(my_job_to_delete)

        if not current_user.is_anonymous:
            my_jobs = get_logic().get_job_mgr().get_all_jobs_by_user(current_user.id)

            for my_temp_task in my_jobs:
                my_tasks.append({"tid": my_temp_task.task_id,
                                 "name": my_temp_task.name,
                                 "details": my_temp_task.started.strftime("%d-%b-%Y %H:%M:%S"),
                                 "status": my_temp_task.status,
                                 "prog": my_temp_task.progress})

        return HomeHandler.PAGE_TASKS, my_tasks

    @staticmethod
    def post_contact(a_request):
        my_thank_you = None
        if a_request.method == 'POST':
            my_thank_you = "Thank You. We will contact you shortly"

            my_source = a_request.form.get("source")
            my_name = a_request.form.get("name")
            my_email = a_request.form.get("email")
            my_details = a_request.form.get("message")

            my_db_session = get_db().session
            my_contact_mgr = ContactMgr(my_db_session)
            my_ret_val, my_message = my_contact_mgr.add_post(my_name, my_email, my_details, my_source)
            if my_ret_val:
                my_contact_mgr.commit()
            else:
                my_thank_you = my_message

        return HomeHandler.PAGE_ABOUT, ContactForm(), my_thank_you
